from duckdb.experimental.spark.exception import ContributionsAcceptedError


class SparkConf:  # noqa: D101
    def __init__(self) -> None:  # noqa: D107
        raise NotImplementedError

    def contains(self, key: str) -> bool:  # noqa: D102
        raise ContributionsAcceptedError

    def get(self, key: str, defaultValue: str | None = None) -> str | None:  # noqa: D102
        raise ContributionsAcceptedError

    def getAll(self) -> list[tuple[str, str]]:  # noqa: D102
        raise ContributionsAcceptedError

    def set(self, key: str, value: str) -> "SparkConf":  # noqa: D102
        raise ContributionsAcceptedError

    def setAll(self, pairs: list[tuple[str, str]]) -> "SparkConf":  # noqa: D102
        raise ContributionsAcceptedError

    def setAppName(self, value: str) -> "SparkConf":  # noqa: D102
        raise ContributionsAcceptedError

    def setExecutorEnv(  # noqa: D102
        self, key: str | None = None, value: str | None = None, pairs: list[tuple[str, str]] | None = None
    ) -> "SparkConf":
        raise ContributionsAcceptedError

    def setIfMissing(self, key: str, value: str) -> "SparkConf":  # noqa: D102
        raise ContributionsAcceptedError

    def setMaster(self, value: str) -> "SparkConf":  # noqa: D102
        raise ContributionsAcceptedError

    def setSparkHome(self, value: str) -> "SparkConf":  # noqa: D102
        raise ContributionsAcceptedError

    def toDebugString(self) -> str:  # noqa: D102
        raise ContributionsAcceptedError


__all__ = ["SparkConf"]



# ruff: noqa: F401
"""The DuckDB Python Package.

This module re-exports the DuckDB C++ extension (`_duckdb`) and provides DuckDB's public API.

Note:
- Some symbols exposed here are implementation details of DuckDB's C++ engine.
- They are kept for backwards compatibility but are not considered stable API.
- Future versions may move them into submodules with deprecation warnings.
"""

from _duckdb import (
    BinderException,
    CaseExpression,
    CatalogException,
    CoalesceOperator,
    ColumnExpression,
    ConnectionException,
    ConstantExpression,
    ConstraintException,
    ConversionException,
    CSVLineTerminator,
    DatabaseError,
    DataError,
    DefaultExpression,
    DependencyException,
    DuckDBPyConnection,
    DuckDBPyRelation,
    Error,
    ExpectedResultType,
    ExplainType,
    Expression,
    FatalException,
    FunctionExpression,
    HTTPException,
    IntegrityError,
    InternalError,
    InternalException,
    InterruptException,
    InvalidInputException,
    InvalidTypeException,
    IOException,
    LambdaExpression,
    NotImplementedException,
    NotSupportedError,
    OperationalError,
    OutOfMemoryException,
    OutOfRangeException,
    ParserException,
    PermissionException,
    ProgrammingError,
    PythonExceptionHandling,
    RenderMode,
    SequenceException,
    SerializationException,
    SQLExpression,
    StarExpression,
    Statement,
    StatementType,
    SyntaxException,
    TransactionException,
    TypeMismatchException,
    Warning,
    __formatted_python_version__,
    __git_revision__,
    __interactive__,
    __jupyter__,
    __standard_vector_size__,
    _clean_default_connection,
    aggregate,
    alias,
    apilevel,
    append,
    array_type,
    arrow,
    begin,
    checkpoint,
    close,
    commit,
    connect,
    create_function,
    cursor,
    decimal_type,
    default_connection,
    description,
    df,
    disable_profiling,
    distinct,
    dtype,
    duplicate,
    enable_profiling,
    enum_type,
    execute,
    executemany,
    extract_statements,
    fetch_arrow_table,
    fetch_df,
    fetch_df_chunk,
    fetch_record_batch,
    fetchall,
    fetchdf,
    fetchmany,
    fetchnumpy,
    fetchone,
    filesystem_is_registered,
    filter,
    from_arrow,
    from_csv_auto,
    from_df,
    from_parquet,
    from_query,
    get_profiling_information,
    get_table_names,
    install_extension,
    interrupt,
    limit,
    list_filesystems,
    list_type,
    load_extension,
    map_type,
    order,
    paramstyle,
    pl,
    project,
    query,
    query_df,
    query_progress,
    read_csv,
    read_json,
    read_parquet,
    register,
    register_filesystem,
    remove_function,
    rollback,
    row_type,
    rowcount,
    set_default_connection,
    sql,
    sqltype,
    string_type,
    struct_type,
    table,
    table_function,
    tf,
    threadsafety,
    to_arrow_reader,
    to_arrow_table,
    token_type,
    tokenize,
    torch,
    type,
    union_type,
    unregister,
    unregister_filesystem,
    values,
    view,
    write_csv,
)

from duckdb._dbapi_type_object import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    DBAPITypeObject,
)
from duckdb._version import (
    __duckdb_version__,
    __version__,
    version,
)
from duckdb.value.constant import (
    BinaryValue,
    BitValue,
    BlobValue,
    BooleanValue,
    DateValue,
    DecimalValue,
    DoubleValue,
    FloatValue,
    HugeIntegerValue,
    IntegerValue,
    IntervalValue,
    ListValue,
    LongValue,
    MapValue,
    NullValue,
    ShortValue,
    StringValue,
    StructValue,
    TimestampMillisecondValue,
    TimestampNanosecondValue,
    TimestampSecondValue,
    TimestampTimeZoneValue,
    TimestampValue,
    TimeTimeZoneValue,
    TimeValue,
    UnionType,
    UnsignedBinaryValue,
    UnsignedHugeIntegerValue,
    UnsignedIntegerValue,
    UnsignedLongValue,
    UnsignedShortValue,
    UUIDValue,
    Value,
)

__all__: list[str] = [
    "BinaryValue",
    "BinderException",
    "BitValue",
    "BlobValue",
    "BooleanValue",
    "CSVLineTerminator",
    "CaseExpression",
    "CatalogException",
    "CoalesceOperator",
    "ColumnExpression",
    "ConnectionException",
    "ConstantExpression",
    "ConstraintException",
    "ConversionException",
    "DataError",
    "DatabaseError",
    "DateValue",
    "DecimalValue",
    "DefaultExpression",
    "DependencyException",
    "DoubleValue",
    "DuckDBPyConnection",
    "DuckDBPyRelation",
    "Error",
    "ExpectedResultType",
    "ExplainType",
    "Expression",
    "FatalException",
    "FloatValue",
    "FunctionExpression",
    "HTTPException",
    "HugeIntegerValue",
    "IOException",
    "IntegerValue",
    "IntegrityError",
    "InternalError",
    "InternalException",
    "InterruptException",
    "IntervalValue",
    "InvalidInputException",
    "InvalidTypeException",
    "LambdaExpression",
    "ListValue",
    "LongValue",
    "MapValue",
    "NotImplementedException",
    "NotSupportedError",
    "NullValue",
    "OperationalError",
    "OutOfMemoryException",
    "OutOfRangeException",
    "ParserException",
    "PermissionException",
    "ProgrammingError",
    "PythonExceptionHandling",
    "RenderMode",
    "SQLExpression",
    "SequenceException",
    "SerializationException",
    "ShortValue",
    "StarExpression",
    "Statement",
    "StatementType",
    "StringValue",
    "StructValue",
    "SyntaxException",
    "TimeTimeZoneValue",
    "TimeValue",
    "TimestampMillisecondValue",
    "TimestampNanosecondValue",
    "TimestampSecondValue",
    "TimestampTimeZoneValue",
    "TimestampValue",
    "TransactionException",
    "TypeMismatchException",
    "UUIDValue",
    "UnionType",
    "UnsignedBinaryValue",
    "UnsignedHugeIntegerValue",
    "UnsignedIntegerValue",
    "UnsignedLongValue",
    "UnsignedShortValue",
    "Value",
    "Warning",
    "__formatted_python_version__",
    "__git_revision__",
    "__interactive__",
    "__jupyter__",
    "__standard_vector_size__",
    "__version__",
    "_clean_default_connection",
    "aggregate",
    "alias",
    "apilevel",
    "append",
    "array_type",
    "arrow",
    "begin",
    "checkpoint",
    "close",
    "commit",
    "connect",
    "create_function",
    "cursor",
    "decimal_type",
    "default_connection",
    "description",
    "df",
    "disable_profiling",
    "distinct",
    "dtype",
    "duplicate",
    "enable_profiling",
    "enum_type",
    "execute",
    "executemany",
    "extract_statements",
    "fetch_arrow_table",
    "fetch_df",
    "fetch_df_chunk",
    "fetch_record_batch",
    "fetchall",
    "fetchdf",
    "fetchmany",
    "fetchnumpy",
    "fetchone",
    "filesystem_is_registered",
    "filter",
    "from_arrow",
    "from_csv_auto",
    "from_df",
    "from_parquet",
    "from_query",
    "get_profiling_information",
    "get_table_names",
    "install_extension",
    "interrupt",
    "limit",
    "list_filesystems",
    "list_type",
    "load_extension",
    "map_type",
    "order",
    "paramstyle",
    "paramstyle",
    "pl",
    "project",
    "query",
    "query_df",
    "query_progress",
    "read_csv",
    "read_json",
    "read_parquet",
    "register",
    "register_filesystem",
    "remove_function",
    "rollback",
    "row_type",
    "rowcount",
    "set_default_connection",
    "sql",
    "sqltype",
    "string_type",
    "struct_type",
    "table",
    "table_function",
    "tf",
    "threadsafety",
    "threadsafety",
    "to_arrow_reader",
    "to_arrow_table",
    "token_type",
    "tokenize",
    "torch",
    "type",
    "union_type",
    "unregister",
    "unregister_filesystem",
    "values",
    "view",
    "write_csv",
]



import uuid
from collections.abc import Callable
from functools import reduce
from keyword import iskeyword
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
    cast,
    overload,
)

import duckdb
from duckdb import ColumnExpression, Expression, StarExpression

from ..errors import PySparkIndexError, PySparkTypeError, PySparkValueError
from .column import Column
from .readwriter import DataFrameWriter
from .type_utils import duckdb_to_spark_schema
from .types import Row, StructType

if TYPE_CHECKING:
    import pyarrow as pa
    from pandas.core.frame import DataFrame as PandasDataFrame

    from ._typing import ColumnOrName
    from .group import GroupedData
    from .session import SparkSession

from duckdb.experimental.spark.sql import functions as spark_sql_functions


class DataFrame:  # noqa: D101
    def __init__(self, relation: duckdb.DuckDBPyRelation, session: "SparkSession") -> None:  # noqa: D107
        self.relation = relation
        self.session = session
        self._schema = None
        if self.relation is not None:
            self._schema = duckdb_to_spark_schema(self.relation.columns, self.relation.types)

    def show(self, **kwargs) -> None:  # noqa: D102
        self.relation.show()

    def toPandas(self) -> "PandasDataFrame":  # noqa: D102
        return self.relation.df()

    def toArrow(self) -> "pa.Table":
        """Returns the contents of this :class:`DataFrame` as PyArrow ``pyarrow.Table``.

        This is only available if PyArrow is installed and available.

        .. versionadded:: 4.0.0

        Notes:
        -----
        This method should only be used if the resulting PyArrow ``pyarrow.Table`` is
        expected to be small, as all the data is loaded into the driver's memory.

        This API is a developer API.

        Examples:
        --------
        >>> df.toArrow()  # doctest: +SKIP
        pyarrow.Table
        age: int64
        name: string
        ----
        age: [[2,5]]
        name: [["Alice","Bob"]]
        """
        return self.relation.to_arrow_table()

    def createOrReplaceTempView(self, name: str) -> None:
        """Creates or replaces a local temporary view with this :class:`DataFrame`.

        The lifetime of this temporary table is tied to the :class:`SparkSession`
        that was used to create this :class:`DataFrame`.

        Parameters
        ----------
        name : str
            Name of the view.

        Examples:
        --------
        Create a local temporary view named 'people'.

        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], schema=["age", "name"])
        >>> df.createOrReplaceTempView("people")

        Replace the local temporary view.

        >>> df2 = df.filter(df.age > 3)
        >>> df2.createOrReplaceTempView("people")
        >>> df3 = spark.sql("SELECT * FROM people")
        >>> sorted(df3.collect()) == sorted(df2.collect())
        True
        >>> spark.catalog.dropTempView("people")
        True

        """
        self.relation.create_view(name, True)

    def createGlobalTempView(self, name: str) -> None:  # noqa: D102
        raise NotImplementedError

    def withColumnRenamed(self, columnName: str, newName: str) -> "DataFrame":  # noqa: D102
        if columnName not in self.relation:
            msg = f"DataFrame does not contain a column named {columnName}"
            raise ValueError(msg)
        cols = []
        for x in self.relation.columns:
            col = ColumnExpression(x)
            if x.casefold() == columnName.casefold():
                col = col.alias(newName)
            cols.append(col)
        rel = self.relation.select(*cols)
        return DataFrame(rel, self.session)

    def withColumn(self, columnName: str, col: Column) -> "DataFrame":  # noqa: D102
        if not isinstance(col, Column):
            raise PySparkTypeError(
                error_class="NOT_COLUMN",
                message_parameters={"arg_name": "col", "arg_type": type(col).__name__},
            )
        if columnName in self.relation:
            # We want to replace the existing column with this new expression
            cols = []
            for x in self.relation.columns:
                if x.casefold() == columnName.casefold():
                    cols.append(col.expr.alias(columnName))
                else:
                    cols.append(ColumnExpression(x))
        else:
            cols = [ColumnExpression(x) for x in self.relation.columns]
            cols.append(col.expr.alias(columnName))
        rel = self.relation.select(*cols)
        return DataFrame(rel, self.session)

    def withColumns(self, *colsMap: dict[str, Column]) -> "DataFrame":
        """Returns a new :class:`DataFrame` by adding multiple columns or replacing the
        existing columns that have the same names.

        The colsMap is a map of column name and column, the column must only refer to attributes
        supplied by this Dataset. It is an error to add columns that refer to some other Dataset.

        .. versionadded:: 3.3.0
           Added support for multiple columns adding

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        colsMap : dict
            a dict of column name and :class:`Column`. Currently, only a single map is supported.

        Returns:
        -------
        :class:`DataFrame`
            DataFrame with new or replaced columns.

        Examples:
        --------
        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], schema=["age", "name"])
        >>> df.withColumns({"age2": df.age + 2, "age3": df.age + 3}).show()
        +---+-----+----+----+
        |age| name|age2|age3|
        +---+-----+----+----+
        |  2|Alice|   4|   5|
        |  5|  Bob|   7|   8|
        +---+-----+----+----+
        """  # noqa: D205
        # Below code is to help enable kwargs in future.
        assert len(colsMap) == 1
        colsMap = colsMap[0]  # type: ignore[assignment]

        if not isinstance(colsMap, dict):
            raise PySparkTypeError(
                error_class="NOT_DICT",
                message_parameters={
                    "arg_name": "colsMap",
                    "arg_type": type(colsMap).__name__,
                },
            )

        column_names = list(colsMap.keys())
        columns = list(colsMap.values())

        # Compute this only once
        column_names_for_comparison = [x.casefold() for x in column_names]

        cols = []
        for x in self.relation.columns:
            if x.casefold() in column_names_for_comparison:
                idx = column_names_for_comparison.index(x)
                # We extract the column name from the originally passed
                # in ones, as the casing might be different than the one
                # in the relation
                col_name = column_names.pop(idx)
                col = columns.pop(idx)
                cols.append(col.expr.alias(col_name))
            else:
                cols.append(ColumnExpression(x))

        # In case anything is remaining, these are new columns
        # that we need to add to the DataFrame
        for col_name, col in zip(column_names, columns, strict=False):
            cols.append(col.expr.alias(col_name))

        rel = self.relation.select(*cols)
        return DataFrame(rel, self.session)

    def withColumnsRenamed(self, colsMap: dict[str, str]) -> "DataFrame":
        """Returns a new :class:`DataFrame` by renaming multiple columns.
        This is a no-op if the schema doesn't contain the given column names.

        .. versionadded:: 3.4.0
           Added support for multiple columns renaming

        Parameters
        ----------
        colsMap : dict
            a dict of existing column names and corresponding desired column names.
            Currently, only a single map is supported.

        Returns:
        -------
        :class:`DataFrame`
            DataFrame with renamed columns.

        See Also:
        --------
        :meth:`withColumnRenamed`

        Notes:
        -----
        Support Spark Connect

        Examples:
        --------
        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], schema=["age", "name"])
        >>> df = df.withColumns({"age2": df.age + 2, "age3": df.age + 3})
        >>> df.withColumnsRenamed({"age2": "age4", "age3": "age5"}).show()
        +---+-----+----+----+
        |age| name|age4|age5|
        +---+-----+----+----+
        |  2|Alice|   4|   5|
        |  5|  Bob|   7|   8|
        +---+-----+----+----+
        """  # noqa: D205
        if not isinstance(colsMap, dict):
            raise PySparkTypeError(
                error_class="NOT_DICT",
                message_parameters={"arg_name": "colsMap", "arg_type": type(colsMap).__name__},
            )

        unknown_columns = set(colsMap.keys()) - set(self.relation.columns)
        if unknown_columns:
            msg = f"DataFrame does not contain column(s): {', '.join(unknown_columns)}"
            raise ValueError(msg)

        # Compute this only once
        old_column_names = list(colsMap.keys())
        old_column_names_for_comparison = [x.casefold() for x in old_column_names]

        cols = []
        for x in self.relation.columns:
            col = ColumnExpression(x)
            if x.casefold() in old_column_names_for_comparison:
                idx = old_column_names.index(x)
                # We extract the column name from the originally passed
                # in ones, as the casing might be different than the one
                # in the relation
                col_name = old_column_names.pop(idx)
                new_col_name = colsMap[col_name]
                col = col.alias(new_col_name)
            cols.append(col)

        rel = self.relation.select(*cols)
        return DataFrame(rel, self.session)

    def transform(self, func: Callable[..., "DataFrame"], *args: Any, **kwargs: Any) -> "DataFrame":  # noqa: ANN401
        """Returns a new :class:`DataFrame`. Concise syntax for chaining custom transformations.

        .. versionadded:: 3.0.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        func : function
            a function that takes and returns a :class:`DataFrame`.
        *args
            Positional arguments to pass to func.

            .. versionadded:: 3.3.0
        **kwargs
            Keyword arguments to pass to func.

            .. versionadded:: 3.3.0

        Returns:
        -------
        :class:`DataFrame`
            Transformed DataFrame.

        Examples:
        --------
        >>> from pyspark.sql.functions import col
        >>> df = spark.createDataFrame([(1, 1.0), (2, 2.0)], ["int", "float"])
        >>> def cast_all_to_int(input_df):
        ...     return input_df.select([col(col_name).cast("int") for col_name in input_df.columns])
        >>> def sort_columns_asc(input_df):
        ...     return input_df.select(*sorted(input_df.columns))
        >>> df.transform(cast_all_to_int).transform(sort_columns_asc).show()
        +-----+---+
        |float|int|
        +-----+---+
        |    1|  1|
        |    2|  2|
        +-----+---+

        >>> def add_n(input_df, n):
        ...     return input_df.select(
        ...         [(col(col_name) + n).alias(col_name) for col_name in input_df.columns]
        ...     )
        >>> df.transform(add_n, 1).transform(add_n, n=10).show()
        +---+-----+
        |int|float|
        +---+-----+
        | 12| 12.0|
        | 13| 13.0|
        +---+-----+
        """
        result = func(self, *args, **kwargs)
        assert isinstance(result, DataFrame), (
            f"Func returned an instance of type [{type(result)}], should have been DataFrame."
        )
        return result

    def sort(self, *cols: str | Column | list[str | Column], **kwargs: Any) -> "DataFrame":  # noqa: ANN401
        """Returns a new :class:`DataFrame` sorted by the specified column(s).

        Parameters
        ----------
        cols : str, list, or :class:`Column`, optional
             list of :class:`Column` or column names to sort by.

        Other Parameters
        ----------------
        ascending : bool or list, optional, default True
            boolean or list of boolean.
            Sort ascending vs. descending. Specify list for multiple sort orders.
            If a list is specified, the length of the list must equal the length of the `cols`.

        Returns:
        -------
        :class:`DataFrame`
            Sorted DataFrame.

        Examples:
        --------
        >>> from pyspark.sql.functions import desc, asc
        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], schema=["age", "name"])

        Sort the DataFrame in ascending order.

        >>> df.sort(asc("age")).show()
        +---+-----+
        |age| name|
        +---+-----+
        |  2|Alice|
        |  5|  Bob|
        +---+-----+

        Sort the DataFrame in descending order.

        >>> df.sort(df.age.desc()).show()
        +---+-----+
        |age| name|
        +---+-----+
        |  5|  Bob|
        |  2|Alice|
        +---+-----+
        >>> df.orderBy(df.age.desc()).show()
        +---+-----+
        |age| name|
        +---+-----+
        |  5|  Bob|
        |  2|Alice|
        +---+-----+
        >>> df.sort("age", ascending=False).show()
        +---+-----+
        |age| name|
        +---+-----+
        |  5|  Bob|
        |  2|Alice|
        +---+-----+

        Specify multiple columns

        >>> df = spark.createDataFrame(
        ...     [(2, "Alice"), (2, "Bob"), (5, "Bob")], schema=["age", "name"]
        ... )
        >>> df.orderBy(desc("age"), "name").show()
        +---+-----+
        |age| name|
        +---+-----+
        |  5|  Bob|
        |  2|Alice|
        |  2|  Bob|
        +---+-----+

        Specify multiple columns for sorting order at `ascending`.

        >>> df.orderBy(["age", "name"], ascending=[False, False]).show()
        +---+-----+
        |age| name|
        +---+-----+
        |  5|  Bob|
        |  2|  Bob|
        |  2|Alice|
        +---+-----+
        """
        if not cols:
            raise PySparkValueError(
                error_class="CANNOT_BE_EMPTY",
                message_parameters={"item": "column"},
            )
        if len(cols) == 1 and isinstance(cols[0], list):
            cols = cols[0]

        columns = []
        for c in cols:
            _c = c
            if isinstance(c, str):
                _c = spark_sql_functions.col(c)
            elif isinstance(c, int) and not isinstance(c, bool):
                # ordinal is 1-based
                if c > 0:
                    _c = self[c - 1]
                # negative ordinal means sort by desc
                elif c < 0:
                    _c = self[-c - 1].desc()
                else:
                    raise PySparkIndexError(
                        error_class="ZERO_INDEX",
                        message_parameters={},
                    )
            columns.append(_c)

        ascending = kwargs.get("ascending", True)

        if isinstance(ascending, (bool, int)):
            if not ascending:
                columns = [c.desc() for c in columns]
        elif isinstance(ascending, list):
            columns = [c if asc else c.desc() for asc, c in zip(ascending, columns, strict=False)]
        else:
            raise PySparkTypeError(
                error_class="NOT_BOOL_OR_LIST",
                message_parameters={"arg_name": "ascending", "arg_type": type(ascending).__name__},
            )

        columns = [spark_sql_functions._to_column_expr(c) for c in columns]
        rel = self.relation.sort(*columns)
        return DataFrame(rel, self.session)

    orderBy = sort

    def head(self, n: int | None = None) -> Row | None | list[Row]:  # noqa: D102
        if n is None:
            rs = self.head(1)
            return rs[0] if rs else None
        return self.take(n)

    first = head

    def take(self, num: int) -> list[Row]:  # noqa: D102
        return self.limit(num).collect()

    def filter(self, condition: "ColumnOrName") -> "DataFrame":
        """Filters rows using the given condition.

        :func:`where` is an alias for :func:`filter`.

        Parameters
        ----------
        condition : :class:`Column` or str
            a :class:`Column` of :class:`types.BooleanType`
            or a string of SQL expressions.

        Returns:
        -------
        :class:`DataFrame`
            Filtered DataFrame.

        Examples:
        --------
        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], schema=["age", "name"])

        Filter by :class:`Column` instances.

        >>> df.filter(df.age > 3).show()
        +---+----+
        |age|name|
        +---+----+
        |  5| Bob|
        +---+----+
        >>> df.where(df.age == 2).show()
        +---+-----+
        |age| name|
        +---+-----+
        |  2|Alice|
        +---+-----+

        Filter by SQL expression in a string.

        >>> df.filter("age > 3").show()
        +---+----+
        |age|name|
        +---+----+
        |  5| Bob|
        +---+----+
        >>> df.where("age = 2").show()
        +---+-----+
        |age| name|
        +---+-----+
        |  2|Alice|
        +---+-----+
        """
        if isinstance(condition, Column):
            cond = condition.expr
        elif isinstance(condition, str):
            cond = condition
        else:
            raise PySparkTypeError(
                error_class="NOT_COLUMN_OR_STR",
                message_parameters={"arg_name": "condition", "arg_type": type(condition).__name__},
            )
        rel = self.relation.filter(cond)
        return DataFrame(rel, self.session)

    where = filter

    def select(self, *cols) -> "DataFrame":  # noqa: D102
        cols = list(cols)
        if len(cols) == 1:
            cols = cols[0]
        if isinstance(cols, list):
            projections = [x.expr if isinstance(x, Column) else ColumnExpression(x) for x in cols]
        else:
            projections = [cols.expr if isinstance(cols, Column) else ColumnExpression(cols)]
        rel = self.relation.select(*projections)
        return DataFrame(rel, self.session)

    @property
    def columns(self) -> list[str]:
        """Returns all column names as a list.

        Examples:
        --------
        >>> df.columns
        ['age', 'name']
        """
        return [f.name for f in self.schema.fields]

    @property
    def dtypes(self) -> list[tuple[str, str]]:
        """Returns all column names and their data types as a list of tuples.

        Returns:
        -------
        list of tuple
            List of tuples, each tuple containing a column name and its data type as strings.

        Examples:
        --------
        >>> df.dtypes
        [('age', 'bigint'), ('name', 'string')]
        """
        return [(f.name, f.dataType.simpleString()) for f in self.schema.fields]

    def _ipython_key_completions_(self) -> list[str]:
        # Provides tab-completion for column names in PySpark DataFrame
        # when accessed in bracket notation, e.g. df['<TAB>]
        return self.columns

    def __dir__(self) -> list[str]:  # noqa: D105
        out = set(super().__dir__())
        out.update(c for c in self.columns if c.isidentifier() and not iskeyword(c))
        return sorted(out)

    def join(
        self,
        other: "DataFrame",
        on: str | list[str] | Column | list[Column] | None = None,
        how: str | None = None,
    ) -> "DataFrame":
        """Joins with another :class:`DataFrame`, using the given join expression.

        Parameters
        ----------
        other : :class:`DataFrame`
            Right side of the join
        on : str, list or :class:`Column`, optional
            a string for the join column name, a list of column names,
            a join expression (Column), or a list of Columns.
            If `on` is a string or a list of strings indicating the name of the join column(s),
            the column(s) must exist on both sides, and this performs an equi-join.
        how : str, optional
            default ``inner``. Must be one of: ``inner``, ``cross``, ``outer``,
            ``full``, ``fullouter``, ``full_outer``, ``left``, ``leftouter``, ``left_outer``,
            ``right``, ``rightouter``, ``right_outer``, ``semi``, ``leftsemi``, ``left_semi``,
            ``anti``, ``leftanti`` and ``left_anti``.

        Returns:
        -------
        :class:`DataFrame`
            Joined DataFrame.

        Examples:
        --------
        The following performs a full outer join between ``df1`` and ``df2``.

        >>> from pyspark.sql import Row
        >>> from pyspark.sql.functions import desc
        >>> df = spark.createDataFrame([(2, "Alice"), (5, "Bob")]).toDF("age", "name")
        >>> df2 = spark.createDataFrame([Row(height=80, name="Tom"), Row(height=85, name="Bob")])
        >>> df3 = spark.createDataFrame([Row(age=2, name="Alice"), Row(age=5, name="Bob")])
        >>> df4 = spark.createDataFrame(
        ...     [
        ...         Row(age=10, height=80, name="Alice"),
        ...         Row(age=5, height=None, name="Bob"),
        ...         Row(age=None, height=None, name="Tom"),
        ...         Row(age=None, height=None, name=None),
        ...     ]
        ... )

        Inner join on columns (default)

        >>> df.join(df2, "name").select(df.name, df2.height).show()
        +----+------+
        |name|height|
        +----+------+
        | Bob|    85|
        +----+------+
        >>> df.join(df4, ["name", "age"]).select(df.name, df.age).show()
        +----+---+
        |name|age|
        +----+---+
        | Bob|  5|
        +----+---+

        Outer join for both DataFrames on the 'name' column.

        >>> df.join(df2, df.name == df2.name, "outer").select(df.name, df2.height).sort(
        ...     desc("name")
        ... ).show()
        +-----+------+
        | name|height|
        +-----+------+
        |  Bob|    85|
        |Alice|  NULL|
        | NULL|    80|
        +-----+------+
        >>> df.join(df2, "name", "outer").select("name", "height").sort(desc("name")).show()
        +-----+------+
        | name|height|
        +-----+------+
        |  Tom|    80|
        |  Bob|    85|
        |Alice|  NULL|
        +-----+------+

        Outer join for both DataFrams with multiple columns.

        >>> df.join(df3, [df.name == df3.name, df.age == df3.age], "outer").select(
        ...     df.name, df3.age
        ... ).show()
        +-----+---+
        | name|age|
        +-----+---+
        |Alice|  2|
        |  Bob|  5|
        +-----+---+
        """
        if on is not None and not isinstance(on, list):
            on = [on]  # type: ignore[assignment]
        if on is not None and not all(isinstance(x, str) for x in on):
            assert isinstance(on, list)
            # Get (or create) the Expressions from the list of Columns
            on = [spark_sql_functions._to_column_expr(x) for x in on]

            # & all the Expressions together to form one Expression
            assert isinstance(on[0], Expression), "on should be Column or list of Column"
            on = reduce(lambda x, y: x.__and__(y), cast("list[Expression]", on))

        if on is None and how is None:
            result = self.relation.join(other.relation)
        else:
            if how is None:
                how = "inner"
            if on is None:
                on = "true"
            elif isinstance(on, list) and all(isinstance(x, str) for x in on):
                # Passed directly through as a list of strings
                on = on
            else:
                on = str(on)
            assert isinstance(how, str), "how should be a string"

            def map_to_recognized_jointype(how: str) -> str:
                known_aliases = {
                    "inner": [],
                    "outer": ["full", "fullouter", "full_outer"],
                    "left": ["leftouter", "left_outer"],
                    "right": ["rightouter", "right_outer"],
                    "anti": ["leftanti", "left_anti"],
                    "semi": ["leftsemi", "left_semi"],
                }
                for type, aliases in known_aliases.items():
                    if how == type or how in aliases:
                        return type
                return how

            how = map_to_recognized_jointype(how)
            result = self.relation.join(other.relation, on, how)
        return DataFrame(result, self.session)

    def crossJoin(self, other: "DataFrame") -> "DataFrame":
        """Returns the cartesian product with another :class:`DataFrame`.

        .. versionadded:: 2.1.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        other : :class:`DataFrame`
            Right side of the cartesian product.

        Returns:
        -------
        :class:`DataFrame`
            Joined DataFrame.

        Examples:
        --------
        >>> from pyspark.sql import Row
        >>> df = spark.createDataFrame([(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])
        >>> df2 = spark.createDataFrame([Row(height=80, name="Tom"), Row(height=85, name="Bob")])
        >>> df.crossJoin(df2.select("height")).select("age", "name", "height").show()
        +---+-----+------+
        |age| name|height|
        +---+-----+------+
        | 14|  Tom|    80|
        | 14|  Tom|    85|
        | 23|Alice|    80|
        | 23|Alice|    85|
        | 16|  Bob|    80|
        | 16|  Bob|    85|
        +---+-----+------+
        """
        return DataFrame(self.relation.cross(other.relation), self.session)

    def alias(self, alias: str) -> "DataFrame":
        """Returns a new :class:`DataFrame` with an alias set.

        Parameters
        ----------
        alias : str
            an alias name to be set for the :class:`DataFrame`.

        Returns:
        -------
        :class:`DataFrame`
            Aliased DataFrame.

        Examples:
        --------
        >>> from pyspark.sql.functions import col, desc
        >>> df = spark.createDataFrame([(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])
        >>> df_as1 = df.alias("df_as1")
        >>> df_as2 = df.alias("df_as2")
        >>> joined_df = df_as1.join(df_as2, col("df_as1.name") == col("df_as2.name"), "inner")
        >>> joined_df.select("df_as1.name", "df_as2.name", "df_as2.age").sort(
        ...     desc("df_as1.name")
        ... ).show()
        +-----+-----+---+
        | name| name|age|
        +-----+-----+---+
        |  Tom|  Tom| 14|
        |  Bob|  Bob| 16|
        |Alice|Alice| 23|
        +-----+-----+---+
        """
        assert isinstance(alias, str), "alias should be a string"
        return DataFrame(self.relation.set_alias(alias), self.session)

    def drop(self, *cols: "ColumnOrName") -> "DataFrame":  # type: ignore[misc]  # noqa: D102
        exclude = []
        for col in cols:
            if isinstance(col, str):
                exclude.append(col)
            elif isinstance(col, Column):
                exclude.append(col.expr.get_name())
            else:
                raise PySparkTypeError(
                    error_class="NOT_COLUMN_OR_STR",
                    message_parameters={"arg_name": "col", "arg_type": type(col).__name__},
                )
        # Filter out the columns that don't exist in the relation
        exclude = [x for x in exclude if x in self.relation.columns]
        expr = StarExpression(exclude=exclude)
        return DataFrame(self.relation.select(expr), self.session)

    def __repr__(self) -> str:  # noqa: D105
        return str(self.relation)

    def limit(self, num: int) -> "DataFrame":
        """Limits the result count to the number specified.

        Parameters
        ----------
        num : int
            Number of records to return. Will return this number of records
            or all records if the DataFrame contains less than this number of records.

        Returns:
        -------
        :class:`DataFrame`
            Subset of the records

        Examples:
        --------
        >>> df = spark.createDataFrame([(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])
        >>> df.limit(1).show()
        +---+----+
        |age|name|
        +---+----+
        | 14| Tom|
        +---+----+
        >>> df.limit(0).show()
        +---+----+
        |age|name|
        +---+----+
        +---+----+
        """
        rel = self.relation.limit(num)
        return DataFrame(rel, self.session)

    def __contains__(self, item: str) -> bool:
        """Check if the :class:`DataFrame` contains a column by the name of `item`."""
        return item in self.relation

    @property
    def schema(self) -> StructType:
        """Returns the schema of this :class:`DataFrame` as a :class:`duckdb.experimental.spark.sql.types.StructType`.

        Examples:
        --------
        >>> df.schema
        StructType([StructField('age', IntegerType(), True),
                    StructField('name', StringType(), True)])
        """
        return self._schema

    @overload
    def __getitem__(self, item: int | str) -> Column: ...

    @overload
    def __getitem__(self, item: Column | list | tuple) -> "DataFrame": ...

    def __getitem__(self, item: int | str | Column | list | tuple) -> Union[Column, "DataFrame"]:
        """Returns the column as a :class:`Column`.

        Examples:
        --------
        >>> df.select(df["age"]).collect()
        [Row(age=2), Row(age=5)]
        >>> df[["name", "age"]].collect()
        [Row(name='Alice', age=2), Row(name='Bob', age=5)]
        >>> df[df.age > 3].collect()
        [Row(age=5, name='Bob')]
        >>> df[df[0] > 3].collect()
        [Row(age=5, name='Bob')]
        """
        if isinstance(item, str):
            return Column(duckdb.ColumnExpression(self.relation.alias, item))
        elif isinstance(item, Column):
            return self.filter(item)
        elif isinstance(item, (list, tuple)):
            return self.select(*item)
        elif isinstance(item, int):
            return spark_sql_functions.col(self._schema[item].name)
        else:
            msg = f"Unexpected item type: {type(item)}"
            raise TypeError(msg)

    def __getattr__(self, name: str) -> Column:
        """Returns the :class:`Column` denoted by ``name``.

        Examples:
        --------
        >>> df.select(df.age).collect()
        [Row(age=2), Row(age=5)]
        """
        if name not in self.relation.columns:
            msg = f"'{self.__class__.__name__}' object has no attribute '{name}'"
            raise AttributeError(msg)
        return Column(duckdb.ColumnExpression(self.relation.alias, name))

    @overload
    def groupBy(self, *cols: "ColumnOrName") -> "GroupedData": ...

    @overload
    def groupBy(self, __cols: list[Column] | list[str]) -> "GroupedData": ...  # noqa: PYI063

    def groupBy(self, *cols: "ColumnOrName") -> "GroupedData":  # type: ignore[misc]
        """Groups the :class:`DataFrame` using the specified columns,
        so we can run aggregation on them. See :class:`GroupedData`
        for all the available aggregate functions.

        :func:`groupby` is an alias for :func:`groupBy`.

        Parameters
        ----------
        cols : list, str or :class:`Column`
            columns to group by.
            Each element should be a column name (string) or an expression (:class:`Column`)
            or list of them.

        Returns:
        -------
        :class:`GroupedData`
            Grouped data by given columns.

        Examples:
        --------
        >>> df = spark.createDataFrame(
        ...     [(2, "Alice"), (2, "Bob"), (2, "Bob"), (5, "Bob")], schema=["age", "name"]
        ... )

        Empty grouping columns triggers a global aggregation.

        >>> df.groupBy().avg().show()
        +--------+
        |avg(age)|
        +--------+
        |    2.75|
        +--------+

        Group-by 'name', and specify a dictionary to calculate the summation of 'age'.

        >>> df.groupBy("name").agg({"age": "sum"}).sort("name").show()
        +-----+--------+
        | name|sum(age)|
        +-----+--------+
        |Alice|       2|
        |  Bob|       9|
        +-----+--------+

        Group-by 'name', and calculate maximum values.

        >>> df.groupBy(df.name).max().sort("name").show()
        +-----+--------+
        | name|max(age)|
        +-----+--------+
        |Alice|       2|
        |  Bob|       5|
        +-----+--------+

        Group-by 'name' and 'age', and calculate the number of rows in each group.

        >>> df.groupBy(["name", df.age]).count().sort("name", "age").show()
        +-----+---+-----+
        | name|age|count|
        +-----+---+-----+
        |Alice|  2|    1|
        |  Bob|  2|    2|
        |  Bob|  5|    1|
        +-----+---+-----+
        """  # noqa: D205
        from .group import GroupedData, Grouping

        columns = cols[0] if len(cols) == 1 and isinstance(cols[0], list) else cols
        return GroupedData(Grouping(*columns), self)

    groupby = groupBy

    @property
    def write(self) -> DataFrameWriter:  # noqa: D102
        return DataFrameWriter(self)

    def printSchema(self, level: int | None = None) -> None:
        """Prints out the schema in the tree format.

        Parameters
        ----------
        level : int, optional
            How many levels to print for nested schemas. Prints all levels by default.

        Examples:
        --------
        >>> df.printSchema()
        root
         |-- age: bigint (nullable = true)
         |-- name: string (nullable = true)
        """
        if level is not None and level < 0:
            raise PySparkValueError(
                error_class="NEGATIVE_VALUE",
                message_parameters={"arg_name": "level", "arg_value": str(level)},
            )
        print(self.schema.treeString(level))

    def union(self, other: "DataFrame") -> "DataFrame":
        """Return a new :class:`DataFrame` containing union of rows in this and another
        :class:`DataFrame`.

        Parameters
        ----------
        other : :class:`DataFrame`
            Another :class:`DataFrame` that needs to be unioned

        Returns:
        -------
        :class:`DataFrame`

        See Also:
        --------
        DataFrame.unionAll

        Notes:
        -----
        This is equivalent to `UNION ALL` in SQL. To do a SQL-style set union
        (that does deduplication of elements), use this function followed by :func:`distinct`.

        Also as standard in SQL, this function resolves columns by position (not by name).

        Examples:
        --------
        >>> df1 = spark.createDataFrame([[1, 2, 3]], ["col0", "col1", "col2"])
        >>> df2 = spark.createDataFrame([[4, 5, 6]], ["col1", "col2", "col0"])
        >>> df1.union(df2).show()
        +----+----+----+
        |col0|col1|col2|
        +----+----+----+
        |   1|   2|   3|
        |   4|   5|   6|
        +----+----+----+
        >>> df1.union(df1).show()
        +----+----+----+
        |col0|col1|col2|
        +----+----+----+
        |   1|   2|   3|
        |   1|   2|   3|
        +----+----+----+
        """  # noqa: D205
        return DataFrame(self.relation.union(other.relation), self.session)

    unionAll = union

    def unionByName(self, other: "DataFrame", allowMissingColumns: bool = False) -> "DataFrame":
        """Returns a new :class:`DataFrame` containing union of rows in this and another :class:`DataFrame`.

        This is different from both `UNION ALL` and `UNION DISTINCT` in SQL. To do a SQL-style set
        union (that does deduplication of elements), use this function followed by :func:`distinct`.

        .. versionadded:: 2.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        other : :class:`DataFrame`
            Another :class:`DataFrame` that needs to be combined.
        allowMissingColumns : bool, optional, default False
           Specify whether to allow missing columns.

           .. versionadded:: 3.1.0

        Returns:
        -------
        :class:`DataFrame`
            Combined DataFrame.

        Examples:
        --------
        The difference between this function and :func:`union` is that this function
        resolves columns by name (not by position):

        >>> df1 = spark.createDataFrame([[1, 2, 3]], ["col0", "col1", "col2"])
        >>> df2 = spark.createDataFrame([[4, 5, 6]], ["col1", "col2", "col0"])
        >>> df1.unionByName(df2).show()
        +----+----+----+
        |col0|col1|col2|
        +----+----+----+
        |   1|   2|   3|
        |   6|   4|   5|
        +----+----+----+

        When the parameter `allowMissingColumns` is ``True``, the set of column names
        in this and other :class:`DataFrame` can differ; missing columns will be filled with null.
        Further, the missing columns of this :class:`DataFrame` will be added at the end
        in the schema of the union result:

        >>> df1 = spark.createDataFrame([[1, 2, 3]], ["col0", "col1", "col2"])
        >>> df2 = spark.createDataFrame([[4, 5, 6]], ["col1", "col2", "col3"])
        >>> df1.unionByName(df2, allowMissingColumns=True).show()
        +----+----+----+----+
        |col0|col1|col2|col3|
        +----+----+----+----+
        |   1|   2|   3|NULL|
        |NULL|   4|   5|   6|
        +----+----+----+----+
        """
        if allowMissingColumns:
            df1 = self.select(
                *self.relation.columns,
                *[
                    spark_sql_functions.lit(None).alias(c)
                    for c in other.relation.columns
                    if c not in self.relation.columns
                ],
            )

            df2 = other.select(
                *[
                    spark_sql_functions.col(c)
                    if c in other.relation.columns
                    else spark_sql_functions.lit(None).alias(c)
                    for c in df1.relation.columns
                ]
            )

            return df1.unionByName(df2, allowMissingColumns=False)
        else:
            other = other.select(*self.relation.columns)

        return DataFrame(self.relation.union(other.relation), self.session)

    def intersect(self, other: "DataFrame") -> "DataFrame":
        """Return a new :class:`DataFrame` containing rows only in
        both this :class:`DataFrame` and another :class:`DataFrame`.
        Note that any duplicates are removed. To preserve duplicates
        use :func:`intersectAll`.

        .. versionadded:: 1.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        other : :class:`DataFrame`
            Another :class:`DataFrame` that needs to be combined.

        Returns:
        -------
        :class:`DataFrame`
            Combined DataFrame.

        Notes:
        -----
        This is equivalent to `INTERSECT` in SQL.

        Examples:
        --------
        >>> df1 = spark.createDataFrame([("a", 1), ("a", 1), ("b", 3), ("c", 4)], ["C1", "C2"])
        >>> df2 = spark.createDataFrame([("a", 1), ("a", 1), ("b", 3)], ["C1", "C2"])
        >>> df1.intersect(df2).sort(df1.C1.desc()).show()
        +---+---+
        | C1| C2|
        +---+---+
        |  b|  3|
        |  a|  1|
        +---+---+
        """  # noqa: D205
        return self.intersectAll(other).drop_duplicates()

    def intersectAll(self, other: "DataFrame") -> "DataFrame":
        """Return a new :class:`DataFrame` containing rows in both this :class:`DataFrame`
        and another :class:`DataFrame` while preserving duplicates.

        This is equivalent to `INTERSECT ALL` in SQL. As standard in SQL, this function
        resolves columns by position (not by name).

        .. versionadded:: 2.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        other : :class:`DataFrame`
            Another :class:`DataFrame` that needs to be combined.

        Returns:
        -------
        :class:`DataFrame`
            Combined DataFrame.

        Examples:
        --------
        >>> df1 = spark.createDataFrame([("a", 1), ("a", 1), ("b", 3), ("c", 4)], ["C1", "C2"])
        >>> df2 = spark.createDataFrame([("a", 1), ("a", 1), ("b", 3)], ["C1", "C2"])
        >>> df1.intersectAll(df2).sort("C1", "C2").show()
        +---+---+
        | C1| C2|
        +---+---+
        |  a|  1|
        |  a|  1|
        |  b|  3|
        +---+---+
        """  # noqa: D205
        return DataFrame(self.relation.intersect(other.relation), self.session)

    def exceptAll(self, other: "DataFrame") -> "DataFrame":
        """Return a new :class:`DataFrame` containing rows in this :class:`DataFrame` but
        not in another :class:`DataFrame` while preserving duplicates.

        This is equivalent to `EXCEPT ALL` in SQL.
        As standard in SQL, this function resolves columns by position (not by name).

        .. versionadded:: 2.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        other : :class:`DataFrame`
            The other :class:`DataFrame` to compare to.

        Returns:
        -------
        :class:`DataFrame`

        Examples:
        --------
        >>> df1 = spark.createDataFrame(
        ...     [("a", 1), ("a", 1), ("a", 1), ("a", 2), ("b", 3), ("c", 4)], ["C1", "C2"]
        ... )
        >>> df2 = spark.createDataFrame([("a", 1), ("b", 3)], ["C1", "C2"])
        >>> df1.exceptAll(df2).show()
        +---+---+
        | C1| C2|
        +---+---+
        |  a|  1|
        |  a|  1|
        |  a|  2|
        |  c|  4|
        +---+---+

        """  # noqa: D205
        return DataFrame(self.relation.except_(other.relation), self.session)

    def dropDuplicates(self, subset: list[str] | None = None) -> "DataFrame":
        """Return a new :class:`DataFrame` with duplicate rows removed,
        optionally only considering certain columns.

        For a static batch :class:`DataFrame`, it just drops duplicate rows. For a streaming
        :class:`DataFrame`, it will keep all data across triggers as intermediate state to drop
        duplicates rows. You can use :func:`withWatermark` to limit how late the duplicate data can
        be and the system will accordingly limit the state. In addition, data older than
        watermark will be dropped to avoid any possibility of duplicates.

        :func:`drop_duplicates` is an alias for :func:`dropDuplicates`.

        Parameters
        ----------
        subset : List of column names, optional
            List of columns to use for duplicate comparison (default All columns).

        Returns:
        -------
        :class:`DataFrame`
            DataFrame without duplicates.

        Examples:
        --------
        >>> from pyspark.sql import Row
        >>> df = spark.createDataFrame(
        ...     [
        ...         Row(name="Alice", age=5, height=80),
        ...         Row(name="Alice", age=5, height=80),
        ...         Row(name="Alice", age=10, height=80),
        ...     ]
        ... )

        Deduplicate the same rows.

        >>> df.dropDuplicates().show()
        +-----+---+------+
        | name|age|height|
        +-----+---+------+
        |Alice|  5|    80|
        |Alice| 10|    80|
        +-----+---+------+

        Deduplicate values on 'name' and 'height' columns.

        >>> df.dropDuplicates(["name", "height"]).show()
        +-----+---+------+
        | name|age|height|
        +-----+---+------+
        |Alice|  5|    80|
        +-----+---+------+
        """  # noqa: D205
        if subset:
            rn_col = f"tmp_col_{uuid.uuid1().hex}"
            subset_str = ", ".join([f'"{c}"' for c in subset])
            window_spec = f"OVER(PARTITION BY {subset_str}) AS {rn_col}"
            df = DataFrame(self.relation.row_number(window_spec, "*"), self.session)
            return df.filter(f"{rn_col} = 1").drop(rn_col)

        return self.distinct()

    drop_duplicates = dropDuplicates

    def distinct(self) -> "DataFrame":
        """Returns a new :class:`DataFrame` containing the distinct rows in this :class:`DataFrame`.

        Returns:
        -------
        :class:`DataFrame`
            DataFrame with distinct records.

        Examples:
        --------
        >>> df = spark.createDataFrame([(14, "Tom"), (23, "Alice"), (23, "Alice")], ["age", "name"])

        Return the number of distinct rows in the :class:`DataFrame`

        >>> df.distinct().count()
        2
        """
        distinct_rel = self.relation.distinct()
        return DataFrame(distinct_rel, self.session)

    def count(self) -> int:
        """Returns the number of rows in this :class:`DataFrame`.

        Returns:
        -------
        int
            Number of rows.

        Examples:
        --------
        >>> df = spark.createDataFrame([(14, "Tom"), (23, "Alice"), (16, "Bob")], ["age", "name"])

        Return the number of rows in the :class:`DataFrame`.

        >>> df.count()
        3
        """
        count_rel = self.relation.count("*")
        return int(count_rel.fetchone()[0])

    def _cast_types(self, *types) -> "DataFrame":
        existing_columns = self.relation.columns
        types_count = len(types)
        assert types_count == len(existing_columns)

        cast_expressions = [
            f"{existing}::{target_type} as {existing}"
            for existing, target_type in zip(existing_columns, types, strict=False)
        ]
        cast_expressions = ", ".join(cast_expressions)
        new_rel = self.relation.project(cast_expressions)
        return DataFrame(new_rel, self.session)

    def toDF(self, *cols) -> "DataFrame":  # noqa: D102
        existing_columns = self.relation.columns
        column_count = len(cols)
        if column_count != len(existing_columns):
            raise PySparkValueError(message="Provided column names and number of columns in the DataFrame don't match")

        existing_columns = [ColumnExpression(x) for x in existing_columns]
        projections = [existing.alias(new) for existing, new in zip(existing_columns, cols, strict=False)]
        new_rel = self.relation.project(*projections)
        return DataFrame(new_rel, self.session)

    def collect(self) -> list[Row]:  # noqa: D102
        columns = self.relation.columns
        result = self.relation.fetchall()

        def construct_row(values: list, names: list[str]) -> Row:
            row = tuple.__new__(Row, list(values))
            row.__fields__ = list(names)
            return row

        rows = [construct_row(x, columns) for x in result]
        return rows

    def cache(self) -> "DataFrame":
        """Persists the :class:`DataFrame` with the default storage level (`MEMORY_AND_DISK_DESER`).

        .. versionadded:: 1.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Notes:
        -----
        The default storage level has changed to `MEMORY_AND_DISK_DESER` to match Scala in 3.0.

        Returns:
        -------
        :class:`DataFrame`
            Cached DataFrame.

        Examples:
        --------
        >>> df = spark.range(1)
        >>> df.cache()
        DataFrame[id: bigint]

        >>> df.explain()
        == Physical Plan ==
        InMemoryTableScan ...
        """
        cached_relation = self.relation.execute()
        return DataFrame(cached_relation, self.session)


__all__ = ["DataFrame"]



class ContributionsAcceptedError(NotImplementedError):
    """This method is not planned to be implemented, if you would like to implement this method
    or show your interest in this method to other members of the community,
    feel free to open up a PR or a Discussion over on https://github.com/duckdb/duckdb.
    """  # noqa: D205

    def __init__(self, message: str | None = None) -> None:  # noqa: D107
        doc = self.__class__.__doc__
        if message:
            doc = message + "\n" + doc
        super().__init__(doc)


__all__ = ["ContributionsAcceptedError"]



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections.abc import Callable
from typing import TYPE_CHECKING, overload

from ..exception import ContributionsAcceptedError
from .column import Column
from .dataframe import DataFrame
from .functions import _to_column_expr
from .types import NumericType

# Only import symbols needed for type checking if something is type checking
if TYPE_CHECKING:
    from ._typing import ColumnOrName
    from .session import SparkSession

__all__ = ["GroupedData", "Grouping"]


def _api_internal(self: "GroupedData", name: str, *cols: str) -> DataFrame:
    expressions = ",".join(list(cols))
    group_by = str(self._grouping) if self._grouping else ""
    projections = self._grouping.get_columns()
    jdf = self._df.relation.apply(
        function_name=name,  # aggregate function
        function_aggr=expressions,  # inputs to aggregate
        group_expr=group_by,  # groups
        projected_columns=projections,  # projections
    )
    return DataFrame(jdf, self.session)


def df_varargs_api(f: Callable[..., DataFrame]) -> Callable[..., DataFrame]:
    def _api(self: "GroupedData", *cols: str) -> DataFrame:
        name = f.__name__
        return _api_internal(self, name, *cols)

    _api.__name__ = f.__name__
    _api.__doc__ = f.__doc__
    return _api


class Grouping:  # noqa: D101
    def __init__(self, *cols: "ColumnOrName", **kwargs) -> None:  # noqa: D107
        self._type = ""
        self._cols = [_to_column_expr(x) for x in cols]
        if "special" in kwargs:
            special = kwargs["special"]
            accepted_special = ["cube", "rollup"]
            assert special in accepted_special
            self._type = special

    def get_columns(self) -> str:  # noqa: D102
        columns = ",".join([str(x) for x in self._cols])
        return columns

    def __str__(self) -> str:  # noqa: D105
        columns = self.get_columns()
        if self._type:
            return self._type + "(" + columns + ")"
        return columns


class GroupedData:
    """A set of methods for aggregations on a :class:`DataFrame`,
    created by :func:`DataFrame.groupBy`.

    """  # noqa: D205

    def __init__(self, grouping: Grouping, df: DataFrame) -> None:  # noqa: D107
        self._grouping = grouping
        self._df = df
        self.session: SparkSession = df.session

    def __repr__(self) -> str:  # noqa: D105
        return str(self._df)

    def count(self) -> DataFrame:
        """Counts the number of records for each group.

        Examples:
        --------
        >>> df = spark.createDataFrame(
        ...     [(2, "Alice"), (3, "Alice"), (5, "Bob"), (10, "Bob")], ["age", "name"]
        ... )
        >>> df.show()
        +---+-----+
        |age| name|
        +---+-----+
        |  2|Alice|
        |  3|Alice|
        |  5|  Bob|
        | 10|  Bob|
        +---+-----+

        Group-by name, and count each group.

        >>> df.groupBy(df.name).count().sort("name").show()
        +-----+-----+
        | name|count|
        +-----+-----+
        |Alice|    2|
        |  Bob|    2|
        +-----+-----+
        """
        return _api_internal(self, "count").withColumnRenamed("count_star()", "count")

    @df_varargs_api
    def mean(self, *cols: str) -> DataFrame:
        """Computes average values for each numeric columns for each group.

        :func:`mean` is an alias for :func:`avg`.

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.
        """

    def avg(self, *cols: str) -> DataFrame:
        """Computes average values for each numeric columns for each group.

        :func:`mean` is an alias for :func:`avg`.

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.

        Examples:
        --------
        >>> df = spark.createDataFrame(
        ...     [(2, "Alice", 80), (3, "Alice", 100), (5, "Bob", 120), (10, "Bob", 140)],
        ...     ["age", "name", "height"],
        ... )
        >>> df.show()
        +---+-----+------+
        |age| name|height|
        +---+-----+------+
        |  2|Alice|    80|
        |  3|Alice|   100|
        |  5|  Bob|   120|
        | 10|  Bob|   140|
        +---+-----+------+

        Group-by name, and calculate the mean of the age in each group.

        >>> df.groupBy("name").avg("age").sort("name").show()
        +-----+--------+
        | name|avg(age)|
        +-----+--------+
        |Alice|     2.5|
        |  Bob|     7.5|
        +-----+--------+

        Calculate the mean of the age and height in all data.

        >>> df.groupBy().avg("age", "height").show()
        +--------+-----------+
        |avg(age)|avg(height)|
        +--------+-----------+
        |     5.0|      110.0|
        +--------+-----------+
        """
        columns = list(cols)
        if len(columns) == 0:
            schema = self._df.schema
            # Take only the numeric types of the relation
            columns: list[str] = [x.name for x in schema.fields if isinstance(x.dataType, NumericType)]
        return _api_internal(self, "avg", *columns)

    @df_varargs_api
    def max(self, *cols: str) -> DataFrame:
        """Computes the max value for each numeric columns for each group.

        Examples:
        --------
        >>> df = spark.createDataFrame(
        ...     [(2, "Alice", 80), (3, "Alice", 100), (5, "Bob", 120), (10, "Bob", 140)],
        ...     ["age", "name", "height"],
        ... )
        >>> df.show()
        +---+-----+------+
        |age| name|height|
        +---+-----+------+
        |  2|Alice|    80|
        |  3|Alice|   100|
        |  5|  Bob|   120|
        | 10|  Bob|   140|
        +---+-----+------+

        Group-by name, and calculate the max of the age in each group.

        >>> df.groupBy("name").max("age").sort("name").show()
        +-----+--------+
        | name|max(age)|
        +-----+--------+
        |Alice|       3|
        |  Bob|      10|
        +-----+--------+

        Calculate the max of the age and height in all data.

        >>> df.groupBy().max("age", "height").show()
        +--------+-----------+
        |max(age)|max(height)|
        +--------+-----------+
        |      10|        140|
        +--------+-----------+
        """

    @df_varargs_api
    def min(self, *cols: str) -> DataFrame:
        """Computes the min value for each numeric column for each group.

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.

        Examples:
        --------
        >>> df = spark.createDataFrame(
        ...     [(2, "Alice", 80), (3, "Alice", 100), (5, "Bob", 120), (10, "Bob", 140)],
        ...     ["age", "name", "height"],
        ... )
        >>> df.show()
        +---+-----+------+
        |age| name|height|
        +---+-----+------+
        |  2|Alice|    80|
        |  3|Alice|   100|
        |  5|  Bob|   120|
        | 10|  Bob|   140|
        +---+-----+------+

        Group-by name, and calculate the min of the age in each group.

        >>> df.groupBy("name").min("age").sort("name").show()
        +-----+--------+
        | name|min(age)|
        +-----+--------+
        |Alice|       2|
        |  Bob|       5|
        +-----+--------+

        Calculate the min of the age and height in all data.

        >>> df.groupBy().min("age", "height").show()
        +--------+-----------+
        |min(age)|min(height)|
        +--------+-----------+
        |       2|         80|
        +--------+-----------+
        """

    @df_varargs_api
    def sum(self, *cols: str) -> DataFrame:
        """Computes the sum for each numeric columns for each group.

        Parameters
        ----------
        cols : str
            column names. Non-numeric columns are ignored.

        Examples:
        --------
        >>> df = spark.createDataFrame(
        ...     [(2, "Alice", 80), (3, "Alice", 100), (5, "Bob", 120), (10, "Bob", 140)],
        ...     ["age", "name", "height"],
        ... )
        >>> df.show()
        +---+-----+------+
        |age| name|height|
        +---+-----+------+
        |  2|Alice|    80|
        |  3|Alice|   100|
        |  5|  Bob|   120|
        | 10|  Bob|   140|
        +---+-----+------+

        Group-by name, and calculate the sum of the age in each group.

        >>> df.groupBy("name").sum("age").sort("name").show()
        +-----+--------+
        | name|sum(age)|
        +-----+--------+
        |Alice|       5|
        |  Bob|      15|
        +-----+--------+

        Calculate the sum of the age and height in all data.

        >>> df.groupBy().sum("age", "height").show()
        +--------+-----------+
        |sum(age)|sum(height)|
        +--------+-----------+
        |      20|        440|
        +--------+-----------+
        """

    @overload
    def agg(self, *exprs: Column) -> DataFrame: ...

    @overload
    def agg(self, __exprs: dict[str, str]) -> DataFrame: ...  # noqa: PYI063

    def agg(self, *exprs: Column | dict[str, str]) -> DataFrame:
        """Compute aggregates and returns the result as a :class:`DataFrame`.

        The available aggregate functions can be:

        1. built-in aggregation functions, such as `avg`, `max`, `min`, `sum`, `count`

        2. group aggregate pandas UDFs, created with :func:`pyspark.sql.functions.pandas_udf`

           .. note:: There is no partial aggregation with group aggregate UDFs, i.e.,
               a full shuffle is required. Also, all the data of a group will be loaded into
               memory, so the user should be aware of the potential OOM risk if data is skewed
               and certain groups are too large to fit in memory.

           .. seealso:: :func:`pyspark.sql.functions.pandas_udf`

        If ``exprs`` is a single :class:`dict` mapping from string to string, then the key
        is the column to perform aggregation on, and the value is the aggregate function.

        Alternatively, ``exprs`` can also be a list of aggregate :class:`Column` expressions.

        .. versionadded:: 1.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        exprs : dict
            a dict mapping from column name (string) to aggregate functions (string),
            or a list of :class:`Column`.

        Notes:
        -----
        Built-in aggregation functions and group aggregate pandas UDFs cannot be mixed
        in a single call to this function.

        Examples:
        --------
        >>> from pyspark.sql import functions as F
        >>> from pyspark.sql.functions import pandas_udf, PandasUDFType
        >>> df = spark.createDataFrame(
        ...     [(2, "Alice"), (3, "Alice"), (5, "Bob"), (10, "Bob")], ["age", "name"]
        ... )
        >>> df.show()
        +---+-----+
        |age| name|
        +---+-----+
        |  2|Alice|
        |  3|Alice|
        |  5|  Bob|
        | 10|  Bob|
        +---+-----+

        Group-by name, and count each group.

        >>> df.groupBy(df.name)
        GroupedData[grouping...: [name...], value: [age: bigint, name: string], type: GroupBy]

        >>> df.groupBy(df.name).agg({"*": "count"}).sort("name").show()
        +-----+--------+
        | name|count(1)|
        +-----+--------+
        |Alice|       2|
        |  Bob|       2|
        +-----+--------+

        Group-by name, and calculate the minimum age.

        >>> df.groupBy(df.name).agg(F.min(df.age)).sort("name").show()
        +-----+--------+
        | name|min(age)|
        +-----+--------+
        |Alice|       2|
        |  Bob|       5|
        +-----+--------+

        Same as above but uses pandas UDF.

        >>> @pandas_udf("int", PandasUDFType.GROUPED_AGG)  # doctest: +SKIP
        ... def min_udf(v):
        ...     return v.min()
        >>> df.groupBy(df.name).agg(min_udf(df.age)).sort("name").show()  # doctest: +SKIP
        +-----+------------+
        | name|min_udf(age)|
        +-----+------------+
        |Alice|           2|
        |  Bob|           5|
        +-----+------------+
        """
        assert exprs, "exprs should not be empty"
        if len(exprs) == 1 and isinstance(exprs[0], dict):
            raise ContributionsAcceptedError
        else:
            # Columns
            assert all(isinstance(c, Column) for c in exprs), "all exprs should be Column"
            expressions = list(self._grouping._cols)
            expressions.extend([x.expr for x in exprs])
            group_by = str(self._grouping)
            rel = self._df.relation.select(*expressions, groups=group_by)
        return DataFrame(rel, self.session)

    # TODO: add 'pivot'  # noqa: TD002, TD003



from collections.abc import (
    Callable,
    Iterable,
)
from typing import TYPE_CHECKING, Any, Union, cast

from ..exception import ContributionsAcceptedError
from .types import DataType

if TYPE_CHECKING:
    from ._typing import DateTimeLiteral, DecimalLiteral, LiteralType

from duckdb import ColumnExpression, ConstantExpression, Expression, FunctionExpression
from duckdb.sqltypes import DuckDBPyType

__all__ = ["Column"]


def _get_expr(x: Union["Column", str]) -> Expression:
    return x.expr if isinstance(x, Column) else ConstantExpression(x)


def _func_op(name: str, doc: str = "") -> Callable[["Column"], "Column"]:
    def _(self: "Column") -> "Column":
        njc = getattr(self.expr, name)()
        return Column(njc)

    _.__doc__ = doc
    return _


def _unary_op(
    name: str,
    doc: str = "unary operator",
) -> Callable[["Column"], "Column"]:
    """Create a method for given unary operator."""

    def _(self: "Column") -> "Column":
        # Call the function identified by 'name' on the internal Expression object
        expr = getattr(self.expr, name)()
        return Column(expr)

    _.__doc__ = doc
    return _


def _bin_op(
    name: str,
    doc: str = "binary operator",
) -> Callable[["Column", Union["Column", "LiteralType", "DecimalLiteral", "DateTimeLiteral"]], "Column"]:
    """Create a method for given binary operator."""

    def _(
        self: "Column",
        other: Union["Column", "LiteralType", "DecimalLiteral", "DateTimeLiteral"],
    ) -> "Column":
        jc = _get_expr(other)
        njc = getattr(self.expr, name)(jc)
        return Column(njc)

    _.__doc__ = doc
    return _


def _bin_func(
    name: str,
    doc: str = "binary function",
) -> Callable[["Column", Union["Column", "LiteralType", "DecimalLiteral", "DateTimeLiteral"]], "Column"]:
    """Create a function expression for the given binary function."""

    def _(
        self: "Column",
        other: Union["Column", "LiteralType", "DecimalLiteral", "DateTimeLiteral"],
    ) -> "Column":
        other = _get_expr(other)
        func = FunctionExpression(name, self.expr, other)
        return Column(func)

    _.__doc__ = doc
    return _


class Column:
    """A column in a DataFrame.

    :class:`Column` instances can be created by::

        # 1. Select a column out of a DataFrame

        df.colName
        df["colName"]

        # 2. Create from an expression
        df.colName + 1
        1 / df.colName

    .. versionadded:: 1.3.0
    """

    def __init__(self, expr: Expression) -> None:  # noqa: D107
        self.expr = expr

    # arithmetic operators
    def __neg__(self) -> "Column":  # noqa: D105
        return Column(-self.expr)

    # `and`, `or`, `not` cannot be overloaded in Python,
    # so use bitwise operators as boolean operators
    __and__ = _bin_op("__and__")
    __or__ = _bin_op("__or__")
    __invert__ = _func_op("__invert__")
    __rand__ = _bin_op("__rand__")
    __ror__ = _bin_op("__ror__")

    __add__ = _bin_op("__add__")

    __sub__ = _bin_op("__sub__")

    __mul__ = _bin_op("__mul__")

    __div__ = _bin_op("__div__")

    __truediv__ = _bin_op("__truediv__")

    __mod__ = _bin_op("__mod__")

    __pow__ = _bin_op("__pow__")

    __radd__ = _bin_op("__radd__")

    __rsub__ = _bin_op("__rsub__")

    __rmul__ = _bin_op("__rmul__")

    __rdiv__ = _bin_op("__rdiv__")

    __rtruediv__ = _bin_op("__rtruediv__")

    __rmod__ = _bin_op("__rmod__")

    __rpow__ = _bin_op("__rpow__")

    def __getitem__(self, k: Any) -> "Column":  # noqa: ANN401
        """An expression that gets an item at position ``ordinal`` out of a list,
        or gets an item by key out of a dict.

        .. versionadded:: 1.3.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        k
            a literal value, or a slice object without step.

        Returns:
        -------
        :class:`Column`
            Column representing the item got by key out of a dict, or substrings sliced by
            the given slice object.

        Examples:
        --------
        >>> df = spark.createDataFrame([("abcedfg", {"key": "value"})], ["l", "d"])
        >>> df.select(df.l[slice(1, 3)], df.d["key"]).show()
        +------------------+------+
        |substring(l, 1, 3)|d[key]|
        +------------------+------+
        |               abc| value|
        +------------------+------+
        """  # noqa: D205
        if isinstance(k, slice):
            raise ContributionsAcceptedError
            # if k.step is not None:
            #    raise ValueError("Using a slice with a step value is not supported")
            # return self.substr(k.start, k.stop)
        else:
            # TODO: this is super hacky  # noqa: TD002, TD003
            expr_str = str(self.expr) + "." + str(k)
            return Column(ColumnExpression(expr_str))

    def __getattr__(self, item: Any) -> "Column":  # noqa: ANN401
        """An expression that gets an item at position ``ordinal`` out of a list,
        or gets an item by key out of a dict.

        Parameters
        ----------
        item
            a literal value.

        Returns:
        -------
        :class:`Column`
            Column representing the item got by key out of a dict.

        Examples:
        --------
        >>> df = spark.createDataFrame([("abcedfg", {"key": "value"})], ["l", "d"])
        >>> df.select(df.d.key).show()
        +------+
        |d[key]|
        +------+
        | value|
        +------+
        """  # noqa: D205
        if item.startswith("__"):
            msg = "Can not access __ (dunder) method"
            raise AttributeError(msg)
        return self[item]

    def alias(self, alias: str) -> "Column":  # noqa: D102
        return Column(self.expr.alias(alias))

    def when(self, condition: "Column", value: Union["Column", str]) -> "Column":  # noqa: D102
        if not isinstance(condition, Column):
            msg = "condition should be a Column"
            raise TypeError(msg)
        v = _get_expr(value)
        expr = self.expr.when(condition.expr, v)
        return Column(expr)

    def otherwise(self, value: Union["Column", str]) -> "Column":  # noqa: D102
        v = _get_expr(value)
        expr = self.expr.otherwise(v)
        return Column(expr)

    def cast(self, dataType: DataType | str) -> "Column":  # noqa: D102
        internal_type = DuckDBPyType(dataType) if isinstance(dataType, str) else dataType.duckdb_type
        return Column(self.expr.cast(internal_type))

    def isin(self, *cols: Iterable[Union["Column", str]] | Union["Column", str]) -> "Column":  # noqa: D102
        if len(cols) == 1 and isinstance(cols[0], (list, set)):
            # Only one argument supplied, it's a list
            cols = cast("tuple", cols[0])

        cols = cast(
            "tuple",
            [_get_expr(c) for c in cols],
        )
        return Column(self.expr.isin(*cols))

    # logistic operators
    def __eq__(  # type: ignore[override]
        self,
        other: Union["Column", "LiteralType", "DecimalLiteral", "DateTimeLiteral"],
    ) -> "Column":
        """Binary function."""
        return Column(self.expr == (_get_expr(other)))

    def __ne__(  # type: ignore[override]
        self,
        other: object,
    ) -> "Column":
        """Binary function."""
        return Column(self.expr != (_get_expr(other)))

    __lt__ = _bin_op("__lt__")

    __le__ = _bin_op("__le__")

    __ge__ = _bin_op("__ge__")

    __gt__ = _bin_op("__gt__")

    # String interrogation methods

    contains = _bin_func("contains")
    rlike = _bin_func("regexp_matches")
    like = _bin_func("~~")
    ilike = _bin_func("~~*")
    startswith = _bin_func("starts_with")
    endswith = _bin_func("suffix")

    # order
    _asc_doc = """
    Returns a sort expression based on the ascending order of the column.
    Examples
    --------
    >>> from pyspark.sql import Row
    >>> df = spark.createDataFrame([('Tom', 80), ('Alice', None)], ["name", "height"])
    >>> df.select(df.name).orderBy(df.name.asc()).collect()
    [Row(name='Alice'), Row(name='Tom')]
    """

    _asc_nulls_first_doc = """
    Returns a sort expression based on ascending order of the column, and null values
    return before non-null values.

    Examples
    --------
    >>> from pyspark.sql import Row
    >>> df = spark.createDataFrame([('Tom', 80), (None, 60), ('Alice', None)], ["name", "height"])
    >>> df.select(df.name).orderBy(df.name.asc_nulls_first()).collect()
    [Row(name=None), Row(name='Alice'), Row(name='Tom')]

    """
    _asc_nulls_last_doc = """
    Returns a sort expression based on ascending order of the column, and null values
    appear after non-null values.

    Examples
    --------
    >>> from pyspark.sql import Row
    >>> df = spark.createDataFrame([('Tom', 80), (None, 60), ('Alice', None)], ["name", "height"])
    >>> df.select(df.name).orderBy(df.name.asc_nulls_last()).collect()
    [Row(name='Alice'), Row(name='Tom'), Row(name=None)]

    """
    _desc_doc = """
    Returns a sort expression based on the descending order of the column.
    Examples
    --------
    >>> from pyspark.sql import Row
    >>> df = spark.createDataFrame([('Tom', 80), ('Alice', None)], ["name", "height"])
    >>> df.select(df.name).orderBy(df.name.desc()).collect()
    [Row(name='Tom'), Row(name='Alice')]
    """
    _desc_nulls_first_doc = """
    Returns a sort expression based on the descending order of the column, and null values
    appear before non-null values.

    Examples
    --------
    >>> from pyspark.sql import Row
    >>> df = spark.createDataFrame([('Tom', 80), (None, 60), ('Alice', None)], ["name", "height"])
    >>> df.select(df.name).orderBy(df.name.desc_nulls_first()).collect()
    [Row(name=None), Row(name='Tom'), Row(name='Alice')]

    """
    _desc_nulls_last_doc = """
    Returns a sort expression based on the descending order of the column, and null values
    appear after non-null values.

    Examples
    --------
    >>> from pyspark.sql import Row
    >>> df = spark.createDataFrame([('Tom', 80), (None, 60), ('Alice', None)], ["name", "height"])
    >>> df.select(df.name).orderBy(df.name.desc_nulls_last()).collect()
    [Row(name='Tom'), Row(name='Alice'), Row(name=None)]
    """

    asc = _unary_op("asc", _asc_doc)
    desc = _unary_op("desc", _desc_doc)
    nulls_first = _unary_op("nulls_first")
    nulls_last = _unary_op("nulls_last")

    def asc_nulls_first(self) -> "Column":  # noqa: D102
        return self.asc().nulls_first()

    def asc_nulls_last(self) -> "Column":  # noqa: D102
        return self.asc().nulls_last()

    def desc_nulls_first(self) -> "Column":  # noqa: D102
        return self.desc().nulls_first()

    def desc_nulls_last(self) -> "Column":  # noqa: D102
        return self.desc().nulls_last()

    def isNull(self) -> "Column":  # noqa: D102
        return Column(self.expr.isnull())

    def isNotNull(self) -> "Column":  # noqa: D102
        return Column(self.expr.isnotnull())



from typing import TYPE_CHECKING, cast

from ..errors import PySparkNotImplementedError, PySparkTypeError
from ..exception import ContributionsAcceptedError
from .types import StructType

PrimitiveType = bool | float | int | str
OptionalPrimitiveType = PrimitiveType | None

if TYPE_CHECKING:
    from duckdb.experimental.spark.sql.dataframe import DataFrame
    from duckdb.experimental.spark.sql.session import SparkSession


class DataFrameWriter:  # noqa: D101
    def __init__(self, dataframe: "DataFrame") -> None:  # noqa: D107
        self.dataframe = dataframe

    def saveAsTable(self, table_name: str) -> None:  # noqa: D102
        relation = self.dataframe.relation
        relation.create(table_name)

    def parquet(  # noqa: D102
        self,
        path: str,
        mode: str | None = None,
        partitionBy: str | list[str] | None = None,
        compression: str | None = None,
    ) -> None:
        relation = self.dataframe.relation
        if mode:
            raise NotImplementedError
        if partitionBy:
            raise NotImplementedError

        relation.write_parquet(path, compression=compression)

    def csv(  # noqa: D102
        self,
        path: str,
        mode: str | None = None,
        compression: str | None = None,
        sep: str | None = None,
        quote: str | None = None,
        escape: str | None = None,
        header: bool | str | None = None,
        nullValue: str | None = None,
        escapeQuotes: bool | str | None = None,
        quoteAll: bool | str | None = None,
        dateFormat: str | None = None,
        timestampFormat: str | None = None,
        ignoreLeadingWhiteSpace: bool | str | None = None,
        ignoreTrailingWhiteSpace: bool | str | None = None,
        charToEscapeQuoteEscaping: str | None = None,
        encoding: str | None = None,
        emptyValue: str | None = None,
        lineSep: str | None = None,
    ) -> None:
        if mode not in (None, "overwrite"):
            raise NotImplementedError
        if escapeQuotes:
            raise NotImplementedError
        if ignoreLeadingWhiteSpace:
            raise NotImplementedError
        if ignoreTrailingWhiteSpace:
            raise NotImplementedError
        if charToEscapeQuoteEscaping:
            raise NotImplementedError
        if emptyValue:
            raise NotImplementedError
        if lineSep:
            raise NotImplementedError
        relation = self.dataframe.relation
        relation.write_csv(
            path,
            sep=sep,
            na_rep=nullValue,
            quotechar=quote,
            compression=compression,
            escapechar=escape,
            header=header if isinstance(header, bool) else header == "True",
            encoding=encoding,
            quoting=quoteAll,
            date_format=dateFormat,
            timestamp_format=timestampFormat,
        )


class DataFrameReader:  # noqa: D101
    def __init__(self, session: "SparkSession") -> None:  # noqa: D107
        self.session = session

    def load(  # noqa: D102
        self,
        path: str | list[str] | None = None,
        format: str | None = None,
        schema: StructType | str | None = None,
        **options: OptionalPrimitiveType,
    ) -> "DataFrame":
        from duckdb.experimental.spark.sql.dataframe import DataFrame

        if not isinstance(path, str):
            raise TypeError
        if options:
            raise ContributionsAcceptedError

        rel = None
        if format:
            format = format.lower()
            if format == "csv" or format == "tsv":
                rel = self.session.conn.read_csv(path)
            elif format == "json":
                rel = self.session.conn.read_json(path)
            elif format == "parquet":
                rel = self.session.conn.read_parquet(path)
            else:
                raise ContributionsAcceptedError
        else:
            rel = self.session.conn.sql(f"select * from {path}")
        df = DataFrame(rel, self.session)
        if schema:
            if not isinstance(schema, StructType):
                raise ContributionsAcceptedError
            schema = cast("StructType", schema)
            types, names = schema.extract_types_and_names()
            df = df._cast_types(types)
            df = df.toDF(names)
        return df

    def csv(  # noqa: D102
        self,
        path: str | list[str],
        schema: StructType | str | None = None,
        sep: str | None = None,
        encoding: str | None = None,
        quote: str | None = None,
        escape: str | None = None,
        comment: str | None = None,
        header: bool | str | None = None,
        inferSchema: bool | str | None = None,
        ignoreLeadingWhiteSpace: bool | str | None = None,
        ignoreTrailingWhiteSpace: bool | str | None = None,
        nullValue: str | None = None,
        nanValue: str | None = None,
        positiveInf: str | None = None,
        negativeInf: str | None = None,
        dateFormat: str | None = None,
        timestampFormat: str | None = None,
        maxColumns: int | str | None = None,
        maxCharsPerColumn: int | str | None = None,
        maxMalformedLogPerPartition: int | str | None = None,
        mode: str | None = None,
        columnNameOfCorruptRecord: str | None = None,
        multiLine: bool | str | None = None,
        charToEscapeQuoteEscaping: str | None = None,
        samplingRatio: float | str | None = None,
        enforceSchema: bool | str | None = None,
        emptyValue: str | None = None,
        locale: str | None = None,
        lineSep: str | None = None,
        pathGlobFilter: bool | str | None = None,
        recursiveFileLookup: bool | str | None = None,
        modifiedBefore: bool | str | None = None,
        modifiedAfter: bool | str | None = None,
        unescapedQuoteHandling: str | None = None,
    ) -> "DataFrame":
        if not isinstance(path, str):
            raise NotImplementedError
        if schema and not isinstance(schema, StructType):
            raise ContributionsAcceptedError
        if comment:
            raise ContributionsAcceptedError
        if inferSchema:
            raise ContributionsAcceptedError
        if ignoreLeadingWhiteSpace:
            raise ContributionsAcceptedError
        if ignoreTrailingWhiteSpace:
            raise ContributionsAcceptedError
        if nanValue:
            raise ConnectionAbortedError
        if positiveInf:
            raise ConnectionAbortedError
        if negativeInf:
            raise ConnectionAbortedError
        if negativeInf:
            raise ConnectionAbortedError
        if maxColumns:
            raise ContributionsAcceptedError
        if maxCharsPerColumn:
            raise ContributionsAcceptedError
        if maxMalformedLogPerPartition:
            raise ContributionsAcceptedError
        if mode:
            raise ContributionsAcceptedError
        if columnNameOfCorruptRecord:
            raise ContributionsAcceptedError
        if multiLine:
            raise ContributionsAcceptedError
        if charToEscapeQuoteEscaping:
            raise ContributionsAcceptedError
        if samplingRatio:
            raise ContributionsAcceptedError
        if enforceSchema:
            raise ContributionsAcceptedError
        if emptyValue:
            raise ContributionsAcceptedError
        if locale:
            raise ContributionsAcceptedError
        if pathGlobFilter:
            raise ContributionsAcceptedError
        if recursiveFileLookup:
            raise ContributionsAcceptedError
        if modifiedBefore:
            raise ContributionsAcceptedError
        if modifiedAfter:
            raise ContributionsAcceptedError
        if unescapedQuoteHandling:
            raise ContributionsAcceptedError
        if lineSep:
            # We have support for custom newline, just needs to be ported to 'read_csv'
            raise NotImplementedError

        dtype = None
        names = None
        if schema:
            schema = cast("StructType", schema)
            dtype, names = schema.extract_types_and_names()

        rel = self.session.conn.read_csv(
            path,
            header=header if isinstance(header, bool) else header == "True",
            sep=sep,
            dtype=dtype,
            na_values=nullValue,
            quotechar=quote,
            escapechar=escape,
            encoding=encoding,
            date_format=dateFormat,
            timestamp_format=timestampFormat,
        )
        from ..sql.dataframe import DataFrame

        df = DataFrame(rel, self.session)
        if names:
            df = df.toDF(*names)
        return df

    def parquet(self, *paths: str, **options: "OptionalPrimitiveType") -> "DataFrame":  # noqa: D102
        input = list(paths)
        if len(input) != 1:
            msg = "Only single paths are supported for now"
            raise NotImplementedError(msg)
        option_amount = len(options.keys())
        if option_amount != 0:
            msg = "Options are not supported"
            raise ContributionsAcceptedError(msg)
        path = input[0]
        rel = self.session.conn.read_parquet(path)
        from ..sql.dataframe import DataFrame

        df = DataFrame(rel, self.session)
        return df

    def json(
        self,
        path: str | list[str],
        schema: StructType | str | None = None,
        primitivesAsString: bool | str | None = None,
        prefersDecimal: bool | str | None = None,
        allowComments: bool | str | None = None,
        allowUnquotedFieldNames: bool | str | None = None,
        allowSingleQuotes: bool | str | None = None,
        allowNumericLeadingZero: bool | str | None = None,
        allowBackslashEscapingAnyCharacter: bool | str | None = None,
        mode: str | None = None,
        columnNameOfCorruptRecord: str | None = None,
        dateFormat: str | None = None,
        timestampFormat: str | None = None,
        multiLine: bool | str | None = None,
        allowUnquotedControlChars: bool | str | None = None,
        lineSep: str | None = None,
        samplingRatio: float | str | None = None,
        dropFieldIfAllNull: bool | str | None = None,
        encoding: str | None = None,
        locale: str | None = None,
        pathGlobFilter: bool | str | None = None,
        recursiveFileLookup: bool | str | None = None,
        modifiedBefore: bool | str | None = None,
        modifiedAfter: bool | str | None = None,
        allowNonNumericNumbers: bool | str | None = None,
    ) -> "DataFrame":
        """Loads JSON files and returns the results as a :class:`DataFrame`.

        `JSON Lines <http://jsonlines.org/>`_ (newline-delimited JSON) is supported by default.
        For JSON (one record per file), set the ``multiLine`` parameter to ``true``.

        If the ``schema`` parameter is not specified, this function goes
        through the input once to determine the input schema.

        .. versionadded:: 1.4.0

        .. versionchanged:: 3.4.0
            Supports Spark Connect.

        Parameters
        ----------
        path : str, list or :class:`RDD`
            string represents path to the JSON dataset, or a list of paths,
            or RDD of Strings storing JSON objects.
        schema : :class:`pyspark.sql.types.StructType` or str, optional
            an optional :class:`pyspark.sql.types.StructType` for the input schema or
            a DDL-formatted string (For example ``col0 INT, col1 DOUBLE``).

        Other Parameters
        ----------------
        Extra options
            For the extra options, refer to
            `Data Source Option <https://spark.apache.org/docs/latest/sql-data-sources-json.html#data-source-option>`_
            for the version you use.

            .. # noqa

        Examples:
        --------
        Write a DataFrame into a JSON file and read it back.

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as d:
        ...     # Write a DataFrame into a JSON file
        ...     spark.createDataFrame([{"age": 100, "name": "Hyukjin Kwon"}]).write.mode(
        ...         "overwrite"
        ...     ).format("json").save(d)
        ...
        ...     # Read the JSON file as a DataFrame.
        ...     spark.read.json(d).show()
        +---+------------+
        |age|        name|
        +---+------------+
        |100|Hyukjin Kwon|
        +---+------------+
        """
        if schema is not None:
            msg = "The 'schema' option is not supported"
            raise ContributionsAcceptedError(msg)
        if primitivesAsString is not None:
            msg = "The 'primitivesAsString' option is not supported"
            raise ContributionsAcceptedError(msg)
        if prefersDecimal is not None:
            msg = "The 'prefersDecimal' option is not supported"
            raise ContributionsAcceptedError(msg)
        if allowComments is not None:
            msg = "The 'allowComments' option is not supported"
            raise ContributionsAcceptedError(msg)
        if allowUnquotedFieldNames is not None:
            msg = "The 'allowUnquotedFieldNames' option is not supported"
            raise ContributionsAcceptedError(msg)
        if allowSingleQuotes is not None:
            msg = "The 'allowSingleQuotes' option is not supported"
            raise ContributionsAcceptedError(msg)
        if allowNumericLeadingZero is not None:
            msg = "The 'allowNumericLeadingZero' option is not supported"
            raise ContributionsAcceptedError(msg)
        if allowBackslashEscapingAnyCharacter is not None:
            msg = "The 'allowBackslashEscapingAnyCharacter' option is not supported"
            raise ContributionsAcceptedError(msg)
        if mode is not None:
            msg = "The 'mode' option is not supported"
            raise ContributionsAcceptedError(msg)
        if columnNameOfCorruptRecord is not None:
            msg = "The 'columnNameOfCorruptRecord' option is not supported"
            raise ContributionsAcceptedError(msg)
        if dateFormat is not None:
            msg = "The 'dateFormat' option is not supported"
            raise ContributionsAcceptedError(msg)
        if timestampFormat is not None:
            msg = "The 'timestampFormat' option is not supported"
            raise ContributionsAcceptedError(msg)
        if multiLine is not None:
            msg = "The 'multiLine' option is not supported"
            raise ContributionsAcceptedError(msg)
        if allowUnquotedControlChars is not None:
            msg = "The 'allowUnquotedControlChars' option is not supported"
            raise ContributionsAcceptedError(msg)
        if lineSep is not None:
            msg = "The 'lineSep' option is not supported"
            raise ContributionsAcceptedError(msg)
        if samplingRatio is not None:
            msg = "The 'samplingRatio' option is not supported"
            raise ContributionsAcceptedError(msg)
        if dropFieldIfAllNull is not None:
            msg = "The 'dropFieldIfAllNull' option is not supported"
            raise ContributionsAcceptedError(msg)
        if encoding is not None:
            msg = "The 'encoding' option is not supported"
            raise ContributionsAcceptedError(msg)
        if locale is not None:
            msg = "The 'locale' option is not supported"
            raise ContributionsAcceptedError(msg)
        if pathGlobFilter is not None:
            msg = "The 'pathGlobFilter' option is not supported"
            raise ContributionsAcceptedError(msg)
        if recursiveFileLookup is not None:
            msg = "The 'recursiveFileLookup' option is not supported"
            raise ContributionsAcceptedError(msg)
        if modifiedBefore is not None:
            msg = "The 'modifiedBefore' option is not supported"
            raise ContributionsAcceptedError(msg)
        if modifiedAfter is not None:
            msg = "The 'modifiedAfter' option is not supported"
            raise ContributionsAcceptedError(msg)
        if allowNonNumericNumbers is not None:
            msg = "The 'allowNonNumericNumbers' option is not supported"
            raise ContributionsAcceptedError(msg)

        if isinstance(path, str):
            path = [path]
        if isinstance(path, list):
            if len(path) == 1:
                rel = self.session.conn.read_json(path[0])
                from .dataframe import DataFrame

                df = DataFrame(rel, self.session)
                return df
            raise PySparkNotImplementedError(message="Only a single path is supported for now")
        else:
            raise PySparkTypeError(
                error_class="NOT_STR_OR_LIST_OF_RDD",
                message_parameters={
                    "arg_name": "path",
                    "arg_type": type(path).__name__,
                },
            )


__all__ = ["DataFrameReader", "DataFrameWriter"]



from typing import TYPE_CHECKING

from .types import StructType

if TYPE_CHECKING:
    from .dataframe import DataFrame
    from .session import SparkSession

PrimitiveType = bool | float | int | str
OptionalPrimitiveType = PrimitiveType | None


class DataStreamWriter:  # noqa: D101
    def __init__(self, dataframe: "DataFrame") -> None:  # noqa: D107
        self.dataframe = dataframe

    def toTable(self, table_name: str) -> None:  # noqa: D102
        # Should we register the dataframe or create a table from the contents?
        raise NotImplementedError


class DataStreamReader:  # noqa: D101
    def __init__(self, session: "SparkSession") -> None:  # noqa: D107
        self.session = session

    def load(  # noqa: D102
        self,
        path: str | None = None,
        format: str | None = None,
        schema: StructType | str | None = None,
        **options: OptionalPrimitiveType,
    ) -> "DataFrame":
        raise NotImplementedError


__all__ = ["DataStreamReader", "DataStreamWriter"]



from __future__ import annotations

from typing import TypeAlias, TYPE_CHECKING, Protocol, Any, TypeVar, Generic, Literal
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from uuid import UUID
from collections.abc import Mapping, Iterator, Sequence, Callable

if TYPE_CHECKING:
    import pyarrow as pa
    from ._expression import Expression
    from ._sqltypes import DuckDBPyType

# Numpy protocols

_S_co = TypeVar("_S_co", bound=tuple[Any, ...], covariant=True)
_D_co = TypeVar("_D_co", covariant=True)

class NPProtocol(Protocol):
    """Base Protocol for numpy objects."""
    @property
    def dtype(self) -> Any: ...
    @property
    def ndim(self) -> int: ...
    def __array__(self, *args: Any, **kwargs: Any) -> Any: ...
    def __array_wrap__(self, *args: Any, **kwargs: Any) -> Any: ...
    @property
    def __array_interface__(self) -> dict[str, Any]: ...
    @property
    def __array_priority__(self) -> float: ...

class NPScalarTypeLike(NPProtocol, Protocol):
    @property
    def itemsize(self) -> int: ...

class NPArrayLike(NPProtocol, Generic[_S_co, _D_co], Protocol):
    """`numpy.ndarray` Protocol.

    This is needed to accept numpy arrays as literals in expressions, without emitting type checker errors about unknown symbol if the user doesn't have `numpy` installed.

    Note:
        Using `np.typing.NDArray` is still the best option for return types.
    """
    def __len__(self) -> int: ...
    def __contains__(self, value: object, /) -> bool: ...
    def __iter__(self) -> Iterator[_D_co]: ...
    def __array_finalize__(self, *args: Any, **kwargs: Any) -> None: ...
    def __getitem__(self, *args: Any, **kwargs: Any) -> Any: ...
    def __setitem__(self, *args: Any, **kwargs: Any) -> None: ...
    @property
    def shape(self) -> _S_co: ...
    @property
    def size(self) -> int: ...

# Expression and values conversions

NumericLiteral: TypeAlias = int | float
"""Python objects that can be converted to a numerical `Expression` or `DuckDBPyType`."""
TemporalLiteral: TypeAlias = date | datetime | time | timedelta
BlobLiteral: TypeAlias = bytes | bytearray
"""Python objects that can be converted to a `BLOB` `ConstantExpression` or `DuckDBPyType`.

Note:
    `bytes` can also be converted to a `BITSTRING`.
"""
ScalarLiteral: TypeAlias = NumericLiteral | BlobLiteral | str | bool
NonNestedLiteral: TypeAlias = ScalarLiteral | TemporalLiteral | UUID | Decimal | memoryview

# NOTE:
# Using `Sequence` and `Mapping` instead of `list | tuple` and `dict` would make the covariance of the element types work.
# Thus, this would allow to avoid the use of `Any` for them.
# However, this would also be incorrect at runtime, since only the 3 aforementioned containers types are accepted.
NestedLiteral: TypeAlias = list[Any] | tuple[Any, ...] | dict[Any, Any] | NPArrayLike[Any, Any]
"""Containers types that can be converted to a nested `ConstantExpression` (e.g. to `ARRAY` or `STRUCT`).

Those types can be arbitrarily nested, as long as their leaf values are `PythonLiteral`."""

PythonLiteral: TypeAlias = NonNestedLiteral | NestedLiteral | None
"""Python objects that can be converted to a `ConstantExpression`."""

IntoExprColumn: TypeAlias = Expression | str
"""Types that are, or can be used as a `ColumnExpression`."""

IntoExpr: TypeAlias = IntoExprColumn | PythonLiteral
"""Any type that can be converted to an `Expression` (or is already one).

See Also:
    https://duckdb.org/docs/stable/clients/python/conversion
"""

IntoValues: TypeAlias = list[PythonLiteral] | tuple[Expression, ...] | Expression | NPArrayLike[Any, Any]
"""Types that can be converted to a table of values."""

# PyType conversions

Builtins: TypeAlias = Literal[
    "bigint",
    "bit",
    "bignum",
    "blob",
    "boolean",
    "date",
    "double",
    "float",
    "geometry",
    "hugeint",
    "integer",
    "interval",
    "smallint",
    "null",
    "time with time zone",
    "time",
    "time_ns",
    "timestamp_ms",
    "timestamp_ns",
    "timestamp_s",
    "timestamp with time zone",
    "timestamp",
    "tinyint",
    "ubigint",
    "uhugeint",
    "uinteger",
    "usmallint",
    "utinyint",
    "uuid",
    "varchar",
    "variant",
]
"""Literals `str` that can be converted into `DuckDBPyType` instances.

Note:
    Passing the same values in uppercase is also accepted. 
    We use lowercase here to be able to reuse this `Literal` in the `DTypeIdentifiers` `Literal`.
"""

NestedIds: TypeAlias = Literal["list", "struct", "array", "enum", "map", "decimal", "union"]
"""Identifiers for nested types in `DuckDBPyType.id`."""

PyTypeIds: TypeAlias = Builtins | NestedIds
"""All possible identifiers for `DuckDBPyType.id`."""

StrIntoPyType: TypeAlias = Builtins | Literal["json"] | str
"""Any `str` that can be converted into a `DuckDBPyType`.

The `DuckDBPyType` not present in the literal values are the composed ones, like `STRUCT` or `DECIMAL`.

Note:
    A `StrEnum` will be handled the same way as a `str`."""

# NOTE:
# the `dict` and `list` types are `Any` due to the same limitation mentioned in `NestedLiteral`.
IntoPyType: TypeAlias = (
    DuckDBPyType
    | StrIntoPyType
    | type[NPScalarTypeLike]
    | type[ScalarLiteral]
    | type[list[Any]]
    | type[dict[Any, Any]]
    | dict[Any, Any]
)
"""All types that can be converted to a `DuckDBPyType`.

They can be arbitrarily nested as long as their leaf values are convertible to `DuckDBPyType`.

See Also:
    https://duckdb.org/docs/stable/clients/python/types
"""

# NOTE: here we keep the covariance "hack" and warn the user in the docstring,
# because otherwise we can just resort to `Any` for the `dict` and `list` types.
IntoFields: TypeAlias = Mapping[str, IntoPyType] | Sequence[IntoPyType]
"""Types that can be converted either into: 

- a nested `DuckDBPyType` (e.g. `STRUCT` or `UNION`)
- a schema for file reads

Warning:
    Only `dict` and `list` containers are accepted at runtime. 
    We use `Mapping` and `Sequence` here to satisfy the covariance of the element types.
"""

# Files related

# NOTE: ideally HiveTypes should also be accepted as a Mapping[str, StrIntoPyType].
ColumnsTypes: TypeAlias = Mapping[str, StrIntoPyType]
HiveTypes: TypeAlias = dict[str, StrIntoPyType]
ParquetFieldIdsType: TypeAlias = Mapping[str, int | ParquetFieldIdsType]

_Auto: TypeAlias = Literal["auto"]
ParquetFieldsOptions: TypeAlias = _Auto | ParquetFieldIdsType
"""Types accepted for the `field_ids` parameter in parquet writing methods."""

CsvEncoding: TypeAlias = Literal["utf-8", "utf-16", "latin-1"] | str
"""Encoding options.

All available options not in the literal values can be seen here:
    https://duckdb.org/docs/stable/core_extensions/encodings
"""
JsonFormat: TypeAlias = _Auto | Literal["unstructured", "newline_delimited", "array"]
JsonRecordOptions: TypeAlias = _Auto | Literal["true", "false"]

# compression kinds

_CompressionOptions: TypeAlias = Literal["gzip", "zstd"]
"""Generally available compression options."""
_None: TypeAlias = Literal["none"]
CsvCompression: TypeAlias = _Auto | _None | _CompressionOptions
JsonCompression: TypeAlias = Literal["auto_detect"] | _None | _CompressionOptions
ParquetCompression: TypeAlias = Literal["uncompressed", "brotli", "snappy", "lz4", "lz4_raw"] | _CompressionOptions

# Other

JoinType: TypeAlias = Literal["inner", "left", "right", "outer", "semi", "anti"]
"""Types of join accepted by `DuckDBPyRelation.join` method."""

ProfilerFormat: TypeAlias = Literal["json", "query_tree", "query_tree_optimizer", "no_output", "html", "graphviz"]
"""Formats available in `get_profiling_information` method/function."""
# TODO: this should be a `Protocol` just like `NPArrayLike`.
ArrowUDF: TypeAlias = Callable[..., pa.Table | pa.Array | pa.ChunkedArray]
"""Type accepted for Python UDFs that return Arrow data."""



from _duckdb._func import ARROW, DEFAULT, NATIVE, SPECIAL, FunctionNullHandling, PythonUDFType  # noqa: D104

__all__ = ["ARROW", "DEFAULT", "NATIVE", "SPECIAL", "FunctionNullHandling", "PythonUDFType"]



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re

from .error_classes import ERROR_CLASSES_MAP


class ErrorClassesReader:
    """A reader to load error information from error_classes.py."""

    def __init__(self) -> None:  # noqa: D107
        self.error_info_map = ERROR_CLASSES_MAP

    def get_error_message(self, error_class: str, message_parameters: dict[str, str]) -> str:
        """Returns the completed error message by applying message parameters to the message template."""
        message_template = self.get_message_template(error_class)
        # Verify message parameters.
        message_parameters_from_template = re.findall("<([a-zA-Z0-9_-]+)>", message_template)
        assert set(message_parameters_from_template) == set(message_parameters), (
            f"Undefined error message parameter for error class: {error_class}. Parameters: {message_parameters}"
        )
        table = str.maketrans("<>", "{}")

        return message_template.translate(table).format(**message_parameters)

    def get_message_template(self, error_class: str) -> str:
        """Returns the message template for corresponding error class from error_classes.py.

        For example,
        when given `error_class` is "EXAMPLE_ERROR_CLASS",
        and corresponding error class in error_classes.py looks like the below:

        .. code-block:: python

            "EXAMPLE_ERROR_CLASS" : {
              "message" : [
                "Problem <A> because of <B>."
              ]
            }

        In this case, this function returns:
        "Problem <A> because of <B>."

        For sub error class, when given `error_class` is "EXAMPLE_ERROR_CLASS.SUB_ERROR_CLASS",
        and corresponding error class in error_classes.py looks like the below:

        .. code-block:: python

            "EXAMPLE_ERROR_CLASS" : {
              "message" : [
                "Problem <A> because of <B>."
              ],
              "sub_class" : {
                "SUB_ERROR_CLASS" : {
                  "message" : [
                    "Do <C> to fix the problem."
                  ]
                }
              }
            }

        In this case, this function returns:
        "Problem <A> because <B>. Do <C> to fix the problem."
        """
        error_classes = error_class.split(".")
        len_error_classes = len(error_classes)
        assert len_error_classes in (1, 2)

        # Generate message template for main error class.
        main_error_class = error_classes[0]
        if main_error_class in self.error_info_map:
            main_error_class_info_map = self.error_info_map[main_error_class]
        else:
            msg = f"Cannot find main error class '{main_error_class}'"
            raise ValueError(msg)

        main_message_template = "\n".join(main_error_class_info_map["message"])

        has_sub_class = len_error_classes == 2

        if not has_sub_class:
            message_template = main_message_template
        else:
            # Generate message template for sub error class if exists.
            sub_error_class = error_classes[1]
            main_error_class_subclass_info_map = main_error_class_info_map["sub_class"]
            if sub_error_class in main_error_class_subclass_info_map:
                sub_error_class_info_map = main_error_class_subclass_info_map[sub_error_class]
            else:
                msg = f"Cannot find sub error class '{sub_error_class}'"
                raise ValueError(msg)

            sub_message_template = "\n".join(sub_error_class_info_map["message"])
            message_template = main_message_template + " " + sub_message_template

        return message_template



"""DuckDB's SQL types."""

from _duckdb._sqltypes import (
    BIGINT,
    BIT,
    BLOB,
    BOOLEAN,
    DATE,
    DOUBLE,
    FLOAT,
    HUGEINT,
    INTEGER,
    INTERVAL,
    SMALLINT,
    SQLNULL,
    TIME,
    TIME_NS,
    TIME_TZ,
    TIMESTAMP,
    TIMESTAMP_MS,
    TIMESTAMP_NS,
    TIMESTAMP_S,
    TIMESTAMP_TZ,
    TINYINT,
    UBIGINT,
    UHUGEINT,
    UINTEGER,
    USMALLINT,
    UTINYINT,
    UUID,
    VARCHAR,
    VARIANT,
    DuckDBPyType,
)

__all__ = [
    "BIGINT",
    "BIT",
    "BLOB",
    "BOOLEAN",
    "DATE",
    "DOUBLE",
    "FLOAT",
    "HUGEINT",
    "INTEGER",
    "INTERVAL",
    "SMALLINT",
    "SQLNULL",
    "TIME",
    "TIMESTAMP",
    "TIMESTAMP_MS",
    "TIMESTAMP_NS",
    "TIMESTAMP_S",
    "TIMESTAMP_TZ",
    "TIME_NS",
    "TIME_TZ",
    "TINYINT",
    "UBIGINT",
    "UHUGEINT",
    "UINTEGER",
    "USMALLINT",
    "UTINYINT",
    "UUID",
    "VARCHAR",
    "VARIANT",
    "DuckDBPyType",
]
